"""
Integration tests of `pandas_openscm.reshaping`
"""

from __future__ import annotations

import numpy as np
import pandas as pd
import pytest

from pandas_openscm.testing import create_test_df


@pytest.mark.parametrize(
    "time_col_name, time_col_name_exp",
    (
        (None, "time"),
        ("time", "time"),
        ("Time", "Time"),
        ("year", "year"),
    ),
)
def test_to_long_data_basic(setup_pandas_accessors, time_col_name, time_col_name_exp):
    kwargs = {}
    if time_col_name is not None:
        kwargs["time_col_name"] = time_col_name_exp

    start = create_test_df(
        variables=[(f"variable_{i}", "Mt") for i in range(5)],
        n_scenarios=3,
        n_runs=6,
        timepoints=np.array([1.0, 2.0, 3.0]),
    )

    res = start.openscm.to_long_data(**kwargs)

    exp = start.melt(ignore_index=False, var_name=time_col_name_exp).reset_index()

    pd.testing.assert_frame_equal(res, exp, check_like=True)


def test_to_long_data_nan_handling(setup_pandas_accessors):
    df = pd.DataFrame(
        [[1, np.nan, 1.2], [2.1, 10.2, np.nan]],
        columns=[2010.0, 2015.0, 2025.0],
        index=pd.MultiIndex.from_tuples(
            [
                ("sa", "v1", "K"),
                ("sb", "v1", "K"),
            ],
            names=["scenario", "variable", "unit"],
        ),
    )

    res = df.openscm.to_long_data()
    # nans are kept
    assert res["value"].isnull().any()

    exp = df.melt(ignore_index=False, var_name="time").reset_index()

    pd.testing.assert_frame_equal(res, exp)


def test_to_long_data_nan_handling_index(setup_pandas_accessors):
    df = pd.DataFrame(
        [[1.1, 0.8, 1.2], [2.1, 10.2, 8.4]],
        columns=[2010.0, 2015.0, 2025.0],
        index=pd.MultiIndex.from_tuples(
            [
                ("sa", np.nan, "K"),
                ("sb", "v1", None),
            ],
            names=["scenario", "variable", "unit"],
        ),
    )

    res = df.openscm.to_long_data()
    # nans are kept
    assert res["variable"].isnull().any()
    assert res["unit"].isnull().any()

    exp = df.melt(ignore_index=False, var_name="time").reset_index()

    pd.testing.assert_frame_equal(res, exp)



"""
Test `pandas_openscm.index_manipulation.set_levels`
"""

from __future__ import annotations

import numpy as np
import pandas as pd
import pytest

from pandas_openscm.index_manipulation import set_index_levels_func, set_levels
from pandas_openscm.testing import convert_to_desired_type

pobj_type = pytest.mark.parametrize(
    "pobj_type",
    ("DataFrame", "Series"),
)
"""
Parameterisation to use to check handling of both DataFrame and Series
"""


@pytest.mark.parametrize(
    "start, levels_to_set, exp",
    (
        pytest.param(
            pd.MultiIndex.from_tuples(
                [
                    ("sa", "va", "kg", 0),
                    ("sb", "vb", "m", 1),
                    ("sa", "va", "kg", 2),
                ],
                names=["scenario", "variable", "unit", "run_id"],
            ),
            {"new_variable": "test"},
            pd.MultiIndex.from_tuples(
                [
                    ("sa", "va", "kg", 0, "test"),
                    ("sb", "vb", "m", 1, "test"),
                    ("sa", "va", "kg", 2, "test"),
                ],
                names=["scenario", "variable", "unit", "run_id", "new_variable"],
            ),
            id="set-single-level",
        ),
        pytest.param(
            pd.MultiIndex.from_tuples(
                [
                    ("sa", "va", "kg", 0),
                    ("sb", "vb", "m", 1),
                    ("sa", "va", "kg", 2),
                ],
                names=["scenario", "variable", "unit", "run_id"],
            ),
            {"new_variable": ["a", "b", "c"]},
            pd.MultiIndex.from_tuples(
                [
                    ("sa", "va", "kg", 0, "a"),
                    ("sb", "vb", "m", 1, "b"),
                    ("sa", "va", "kg", 2, "c"),
                ],
                names=["scenario", "variable", "unit", "run_id", "new_variable"],
            ),
            id="set-multiple-values",
        ),
        pytest.param(
            pd.MultiIndex.from_tuples(
                [
                    ("sa", "va", "kg", 5),
                    ("sb", "vb", "m", 1),
                    ("sa", "va", "kg", 2),
                ],
                names=["scenario", "variable", "unit", "run_id"],
            ),
            {"variable": ["a", "b", "c"]},
            pd.MultiIndex.from_tuples(
                [
                    ("sa", "a", "kg", 5),
                    ("sb", "b", "m", 1),
                    ("sa", "c", "kg", 2),
                ],
                names=["scenario", "variable", "unit", "run_id"],
            ),
            id="replace-existing-level",
        ),
        pytest.param(
            pd.MultiIndex.from_tuples(
                [
                    ("sa", "va", "kg", 5),
                    ("sb", "vb", "m", 1),
                    ("sa", "va", "kg", 2),
                ],
                names=["scenario", "variable", "unit", "run_id"],
            ),
            {"new_variable": ["a", "b", "c"], "another_new_variable": [0, 0, 0]},
            pd.MultiIndex.from_tuples(
                [
                    ("sa", "va", "kg", 5, "a", 0),
                    ("sb", "vb", "m", 1, "b", 0),
                    ("sa", "va", "kg", 2, "c", 0),
                ],
                names=[
                    "scenario",
                    "variable",
                    "unit",
                    "run_id",
                    "new_variable",
                    "another_new_variable",
                ],
            ),
            id="add-multiple-new-levels",
        ),
        pytest.param(
            pd.MultiIndex.from_tuples(
                [
                    ("sa", "va", "kg", 5),
                    ("sb", "vb", "m", 1),
                    ("sa", "va", "kg", 2),
                ],
                names=["scenario", "variable", "unit", "run_id"],
            ),
            {
                "scenario": ["a", "b", "c"],
                "extra_variable": [1, 1, 1],
                "single_value_varible": "single_value",
            },
            pd.MultiIndex.from_tuples(
                [
                    ("a", "va", "kg", 5, 1, "single_value"),
                    ("b", "vb", "m", 1, 1, "single_value"),
                    ("c", "va", "kg", 2, 1, "single_value"),
                ],
                names=[
                    "scenario",
                    "variable",
                    "unit",
                    "run_id",
                    "extra_variable",
                    "single_value_varible",
                ],
            ),
            id="replace-existing-level-and-add-new-levels",
        ),
        pytest.param(
            pd.MultiIndex.from_tuples(
                [
                    ("sa", "va", "kg", 5),
                    ("sb", "vb", "m", 1),
                    ("sa", "va", "kg", 2),
                ],
                names=["scenario", "variable", "unit", "run_id"],
            ),
            {"scenarioABC": ["aa", "bb", "cc"], "variable": "v"},
            pd.MultiIndex.from_tuples(
                [
                    ("sa", "v", "kg", 5, "aa"),
                    ("sb", "v", "m", 1, "bb"),
                    ("sa", "v", "kg", 2, "cc"),
                ],
                names=["scenario", "variable", "unit", "run_id", "scenarioABC"],
            ),
            id="add-new-level-replace-level-with-single-value",
        ),
    ),
)
def test_set_levels(start, levels_to_set, exp):
    res = set_levels(start, levels_to_set=levels_to_set)

    pd.testing.assert_index_equal(res, exp)


@pobj_type
def test_set_levels_with_a_dataframe(pobj_type):
    start = pd.MultiIndex.from_tuples(
        [
            ("sa", "va", "kg", 0),
            ("sb", "vb", "m", 1),
            ("sa", "va", "kg", 2),
            ("sa", "vb", "kg", -2),
        ],
        names=["scenario", "variable", "unit", "run_id"],
    )
    start_pobj = convert_to_desired_type(
        pd.DataFrame(
            np.zeros((start.shape[0], 3)), columns=[2010, 2020, 2030], index=start
        ),
        pobj_type,
    )

    res = set_index_levels_func(start_pobj, levels_to_set={"new_variable": "test"})

    exp = pd.MultiIndex.from_tuples(
        [
            ("sa", "va", "kg", 0, "test"),
            ("sb", "vb", "m", 1, "test"),
            ("sa", "va", "kg", 2, "test"),
            ("sa", "vb", "kg", -2, "test"),
        ],
        names=["scenario", "variable", "unit", "run_id", "new_variable"],
    )

    pd.testing.assert_index_equal(res.index, exp)


@pobj_type
def test_set_levels_raises_type_error(pobj_type):
    start = pd.DataFrame(
        np.arange(2 * 4).reshape((4, 2)),
        columns=[2010, 2020],
    )
    start = convert_to_desired_type(start, pobj_type)

    levels_to_set = {"new_variable": "test"}

    with pytest.raises(TypeError):
        set_index_levels_func(start, levels_to_set=levels_to_set)


def test_set_levels_raises_value_error():
    start = pd.MultiIndex.from_tuples(
        [
            ("sa", "va", "kg", 0),
            ("sb", "vb", "m", 1),
            ("sa", "va", "kg", 2),
        ],
        names=["scenario", "variable", "unit", "run_id"],
    )

    levels_to_set = {"new_variable": ["a", "b", "c", "d"]}

    with pytest.raises(
        ValueError,
        match="Length of values for level 'new_variable' "
        "does not match index length: 4 != 3",
    ):
        set_levels(start, levels_to_set=levels_to_set)


def test_accessor_df(setup_pandas_accessors):
    start = pd.DataFrame(
        np.arange(2 * 4).reshape((4, 2)),
        columns=[2010, 2020],
        index=pd.MultiIndex.from_tuples(
            [
                ("sa", "va", "kg", 0),
                ("sb", "vb", "m", -1),
                ("sa", "va", "kg", -2),
                ("sa", "vb", "kg", 2),
            ],
            names=["scenario", "variable", "unit", "run_id"],
        ),
    )

    levels_to_set = {
        "model_id": "674",
        "unit": ["t", "km", "g", "kg"],
        "scenario": 1,
    }

    exp = pd.DataFrame(
        start.values,
        columns=start.columns,
        index=pd.MultiIndex.from_tuples(
            [
                (1, "va", "t", 0, "674"),
                (1, "vb", "km", -1, "674"),
                (1, "va", "g", -2, "674"),
                (1, "vb", "kg", 2, "674"),
            ],
            names=["scenario", "variable", "unit", "run_id", "model_id"],
        ),
    )

    res = start.openscm.set_index_levels(levels_to_set=levels_to_set)
    pd.testing.assert_frame_equal(res, exp)

    # Test function too
    res = set_index_levels_func(start, levels_to_set=levels_to_set)
    pd.testing.assert_frame_equal(res, exp)


def test_accessor_series(setup_pandas_accessors):
    start = pd.Series(
        np.arange(4),
        index=pd.MultiIndex.from_tuples(
            [
                ("sa", "va", "kg", 0),
                ("sb", "vb", "m", -1),
                ("sa", "va", "kg", -2),
                ("sa", "vb", "kg", 2),
            ],
            names=["scenario", "variable", "unit", "run_id"],
        ),
    )

    levels_to_set = {
        "model_id": "674",
        "unit": ["t", "km", "g", "kg"],
        "scenario": 1,
    }

    exp = pd.Series(
        start.values,
        index=pd.MultiIndex.from_tuples(
            [
                (1, "va", "t", 0, "674"),
                (1, "vb", "km", -1, "674"),
                (1, "va", "g", -2, "674"),
                (1, "vb", "kg", 2, "674"),
            ],
            names=["scenario", "variable", "unit", "run_id", "model_id"],
        ),
    )

    res = start.openscm.set_index_levels(levels_to_set=levels_to_set)
    pd.testing.assert_series_equal(res, exp)

    # Test function too
    res = set_index_levels_func(start, levels_to_set=levels_to_set)
    pd.testing.assert_series_equal(res, exp)



"""
Tests of moving the database
"""

from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from pandas_openscm.db import (
    CSVDataBackend,
    CSVIndexBackend,
    FeatherDataBackend,
    FeatherIndexBackend,
    OpenSCMDB,
    netCDFDataBackend,
    netCDFIndexBackend,
)
from pandas_openscm.testing import assert_frame_alike

pytest.importorskip("filelock")


@pytest.mark.parametrize(
    "backend_data, backend_index",
    (
        pytest.param(
            FeatherDataBackend(),
            FeatherIndexBackend(),
            id="feather",
        ),
        pytest.param(
            netCDFDataBackend(),
            netCDFIndexBackend(),
            id="netCDF",
        ),
        pytest.param(
            CSVDataBackend(),
            CSVIndexBackend(),
            id="csv",
        ),
    ),
)
@pytest.mark.parametrize("provide_backend_data_to_class_method", (True, False))
@pytest.mark.parametrize("provide_backend_index_to_class_method", (True, False))
def test_move_db(  # noqa: PLR0913
    provide_backend_index_to_class_method,
    provide_backend_data_to_class_method,
    backend_data,
    backend_index,
    tmpdir,
    setup_pandas_accessors,
):
    initial_db_dir = Path(tmpdir) / "initial"
    other_db_dir = Path(tmpdir) / "other"
    tar_archive = Path(tmpdir) / "tar_archive.tar.gz"

    db = OpenSCMDB(
        db_dir=initial_db_dir,
        backend_data=backend_data,
        backend_index=backend_index,
    )

    df_timeseries_like = pd.DataFrame(
        np.arange(12).reshape(4, 3),
        columns=[2010, 2015, 2025],
        index=pd.MultiIndex.from_tuples(
            [
                ("scenario_a", "climate_model_a", "Temperature", "K"),
                ("scenario_b", "climate_model_a", "Temperature", "K"),
                ("scenario_b", "climate_model_b", "Temperature", "K"),
                ("scenario_b", "climate_model_b", "Ocean Heat Uptake", "J"),
            ],
            names=["scenario", "climate_model", "variable", "unit"],
        ),
    )

    db.save(df_timeseries_like, groupby=["scenario", "variable"])

    # Create a tar archive (returns the archive path, even though it's also an input)
    tar_archive = db.to_gzipped_tar_archive(tar_archive)

    # Expand elsewhere
    from_gzipped_tar_archive_kwargs = {}
    if provide_backend_data_to_class_method:
        from_gzipped_tar_archive_kwargs["backend_data"] = backend_data

    if provide_backend_index_to_class_method:
        from_gzipped_tar_archive_kwargs["backend_index"] = backend_index

    db_other = OpenSCMDB.from_gzipped_tar_archive(
        tar_archive, db_dir=other_db_dir, **from_gzipped_tar_archive_kwargs
    )

    # Delete the original
    db.delete()

    assert_frame_alike(df_timeseries_like, db_other.load(out_columns_type=int))

    locator = pd.Index(["scenario_b"], name="scenario")
    assert_frame_alike(
        df_timeseries_like.openscm.mi_loc(locator),
        db_other.load(locator, out_columns_type=int),
    )



"""
Accessor for [pd.DataFrame][pandas.DataFrame]
"""

from __future__ import annotations

from collections.abc import Collection, Mapping
from typing import TYPE_CHECKING, Any, Callable

import pandas as pd

from pandas_openscm.grouping import (
    fix_index_name_after_groupby_quantile,
    groupby_except,
)
from pandas_openscm.index_manipulation import (
    convert_index_to_category_index,
    ensure_index_is_multiindex,
    set_index_levels_func,
    update_index_levels_from_other_func,
    update_index_levels_func,
)
from pandas_openscm.indexing import mi_loc
from pandas_openscm.plotting import (
    create_legend_default,
    plot_plume_after_calculating_quantiles_func,
    plot_plume_func,
)
from pandas_openscm.reshaping import ts_to_long_data
from pandas_openscm.unit_conversion import convert_unit, convert_unit_like

if TYPE_CHECKING:
    import matplotlib
    import pandas_indexing as pix
    import pint

    from pandas_openscm.plotting import (
        PALETTE_LIKE,
        QUANTILES_PLUMES_LIKE,
    )


class PandasDataFrameOpenSCMAccessor:
    """
    [pd.DataFrame][pandas.DataFrame] accessor

    For details, see
    [pandas' docs](https://pandas.pydata.org/docs/development/extending.html#registering-custom-accessors).
    """

    def __init__(self, df: pd.DataFrame):
        """
        Initialise

        Parameters
        ----------
        df
            [pd.DataFrame][pandas.DataFrame] to use via the accessor
        """
        # It is possible to validate here.
        # However, it's probably better to do validation closer to the data use.
        self._df = df

    def convert_unit(
        self,
        desired_units: str | Mapping[str, str] | pd.Series[str],
        unit_level: str = "unit",
        ur: pint.facets.PlainRegistry | None = None,
    ) -> pd.DataFrame:
        """
        Convert units

        This uses [convert_unit_from_target_series][pandas_openscm.unit_conversion.].
        If you want to understand the details of how the conversion works,
        see that function's docstring.

        Parameters
        ----------
        desired_units
            Desired unit(s) for `df`

            If this is a string,
            we attempt to convert all timeseries to the given unit.

            If this is a mapping,
            we convert the given units to the target units.
            Be careful using this form - you need to be certain of the units.
            If any of your keys don't match the existing units
            (even by a single whitespace character)
            then the unit conversion will not happen.

            If this is a [pd.Series][pandas.Series],
            then it will be passed to
            [convert_unit_from_target_series][pandas_openscm.unit_conversion.]
            after filling any rows in the [pd.DataFrame][pandas.DataFrame]
            that are not in `desired_units`
            with the existing unit (i.e. unspecified rows are not converted).

            For further details, see the examples
            in [convert_unit][pandas_openscm.unit_conversion.].

        unit_level
            Level in the index which holds unit information

            Passed to
            [convert_unit_from_target_series][pandas_openscm.unit_conversion.].

        ur
            Unit registry to use for the conversion.

            Passed to
            [convert_unit_from_target_series][pandas_openscm.unit_conversion.].

        Returns
        -------
        :
            Data with converted units
        """
        return convert_unit(
            self._df, desired_units=desired_units, unit_level=unit_level, ur=ur
        )

    def convert_unit_like(
        self,
        target: pd.DataFrame | pd.Series[Any],
        unit_level: str = "unit",
        target_unit_level: str | None = None,
        ur: pint.facets.PlainRegistry | None = None,
    ) -> pd.DataFrame:
        """
        Convert units to match another [pd.DataFrame][pandas.DataFrame]

        For further details, see the examples
        in [convert_unit_like][pandas_openscm.unit_conversion.].

        This is essentially a helper for
        [convert_unit_from_target_series][pandas_openscm.unit_conversion.].
        It implements one set of logic for extracting desired units
        and tries to be clever, handling differences in index levels
        between the data and `target` sensibly wherever possible.

        If you want behaviour other than what is implemented here,
        use [convert_unit_from_target_series][pandas_openscm.unit_conversion.] directly.

        Parameters
        ----------
        target
            Supported [pandas][] object whose units should be matched

        unit_level
            Level in the data's index which holds unit information

        target_unit_level
            Level in `target`'s index which holds unit information

            If not supplied, we use `df_unit_level`.

        ur
            Unit registry to use for the conversion.

            Passed to
            [convert_unit_from_target_series][pandas_openscm.unit_conversion.].

        Returns
        -------
        :
            Data with converted units
        """
        return convert_unit_like(
            self._df,
            target=target,
            unit_level=unit_level,
            target_unit_level=target_unit_level,
            ur=ur,
        )

    def ensure_index_is_multiindex(self, copy: bool = True) -> pd.DataFrame:
        """
        Ensure that the index is a [pd.MultiIndex][pandas.MultiIndex]

        Parameters
        ----------
        copy
            Whether to copy `df` before manipulating the index name

        Returns
        -------
        :
            `df` with a [pd.MultiIndex][pandas.MultiIndex]

            If the index was already a [pd.MultiIndex][pandas.MultiIndex],
            this is a no-op (although the value of copy is respected).
        """
        return ensure_index_is_multiindex(self._df, copy=copy)

    def eiim(self, copy: bool = True) -> pd.DataFrame:
        """
        Ensure that the index is a [pd.MultiIndex][pandas.MultiIndex]

        Alias for [ensure_index_is_multiindex][pandas_openscm.index_manipulation.]

        Parameters
        ----------
        copy
            Whether to copy `df` before manipulating the index name

        Returns
        -------
        :
            `df` with a [pd.MultiIndex][pandas.MultiIndex]

            If the index was already a [pd.MultiIndex][pandas.MultiIndex],
            this is a no-op (although the value of copy is respected).
        """
        return self.ensure_index_is_multiindex(copy=copy)

    def fix_index_name_after_groupby_quantile(
        self, new_name: str = "quantile", copy: bool = False
    ) -> pd.DataFrame:
        """
        Fix the index name after performing a `groupby(...).quantile(...)` operation

        By default, pandas doesn't assign a name to the quantile level
        when doing an operation of the form given above.
        This fixes this, but it does assume
        that the quantile level is the only unnamed level in the index.

        Parameters
        ----------
        new_name
            New name to give to the quantile column

        copy
            Whether to copy `df` before manipulating the index name

        Returns
        -------
        :
            `df`, with the last level in its index renamed to `new_name`.
        """
        return fix_index_name_after_groupby_quantile(
            self._df, new_name=new_name, copy=copy
        )

    def groupby_except(
        self, non_groupers: str | list[str], observed: bool = True
    ) -> pd.core.groupby.generic.DataFrameGroupBy[Any]:
        """
        Group by all index levels except specified levels

        This is the inverse of [pd.DataFrame.groupby][pandas.DataFrame.groupby].

        Parameters
        ----------
        non_groupers
            Columns to exclude from the grouping

        observed
            Whether to only return observed combinations or not

        Returns
        -------
        :
            The [pd.DataFrame][pandas.DataFrame],
            grouped by all columns except `non_groupers`.
        """
        return groupby_except(df=self._df, non_groupers=non_groupers, observed=observed)

    def mi_loc(
        self,
        locator: pd.Index[Any] | pd.MultiIndex | pix.selectors.Selector,
    ) -> pd.DataFrame:
        """
        Select data, being slightly smarter than the default [pandas.DataFrame.loc][].

        Parameters
        ----------
        locator
            Locator to apply

            If this is a multi-index, we use
            [multi_index_lookup][pandas_openscm.indexing.] to ensure correct alignment.

            If this is an index that has a name,
            we use the name to ensure correct alignment.

        Returns
        -------
        :
            Selected data

        Notes
        -----
        If you have [pandas_indexing][] installed,
        you can get the same (perhaps even better) functionality
        using something like the following instead

        ```python
        ...
        pandas_obj.loc[pandas_indexing.isin(locator)]
        ...
        ```
        """
        return mi_loc(self._df, locator)

    def plot_plume(  # noqa: PLR0913
        self,
        quantiles_plumes: QUANTILES_PLUMES_LIKE,
        ax: matplotlib.axes.Axes | None = None,
        *,
        quantile_var: str = "quantile",
        quantile_var_label: str | None = None,
        quantile_legend_round: int = 3,
        hue_var: str = "scenario",
        hue_var_label: str | None = None,
        palette: PALETTE_LIKE[Any] | None = None,
        warn_on_palette_value_missing: bool = True,
        style_var: str = "variable",
        style_var_label: str | None = None,
        dashes: dict[Any, str | tuple[float, tuple[float, ...]]] | None = None,
        warn_on_dashes_value_missing: bool = True,
        linewidth: float = 2.0,
        unit_var: str = "unit",
        unit_aware: bool | pint.facets.PlainRegistry = False,
        time_units: str | None = None,
        x_label: str | None = "time",
        y_label: str | bool | None = True,
        warn_infer_y_label_with_multi_unit: bool = True,
        create_legend: Callable[
            [matplotlib.axes.Axes, list[matplotlib.artist.Artist]], None
        ] = create_legend_default,
        observed: bool = True,
    ) -> matplotlib.axes.Axes:
        """
        Plot a plume plot

        Parameters
        ----------
        quantiles_plumes
            Quantiles to plot in each plume.

            If the first element of each tuple is a tuple,
            a plume is plotted between the given quantiles.
            Otherwise, if the first element is a plain float,
            a line is plotted for the given quantile.

        ax
            Axes on which to plot.

            If not supplied, a new axes is created.

        quantile_var
            Variable/column in the multi-index which stores information
            about the quantile that each timeseries represents.

        quantile_var_label
            Label to use as the header for the quantile section in the legend

        quantile_legend_round
            Rounding to apply to quantile values when creating the legend

        hue_var
            Variable to use for grouping data into different colour groups

        hue_var_label
            Label to use as the header for the hue/colour section in the legend

        palette
            Colour to use for the different groups in the data.

            If any groups are not included in `palette`,
            they are auto-filled.

        warn_on_palette_value_missing
            Should a warning be emitted if there are values missing from `palette`?

        style_var
            Variable to use for grouping data into different (line)style groups

        style_var_label
            Label to use as the header for the style section in the legend

        dashes
            Dash/linestyle to use for the different groups in the data.

            If any groups are not included in `dashes`,
            they are auto-filled.

        warn_on_dashes_value_missing
            Should a warning be emitted if there are values missing from `dashes`?

        linewidth
            Width to use for plotting lines.

        unit_var
            Variable/column in the multi-index which stores information
            about the unit of each timeseries.

        unit_aware
            Should the plot be done in a unit-aware way?

            If `True`, we use the default application registry
            (retrieved with [pint.get_application_registry][]).
            Otherwise, a [pint.facets.PlainRegistry][] can be supplied and will be used.

            For details, see matplotlib and pint support plotting with units
            ([stable docs](https://pint.readthedocs.io/en/stable/user/plotting.html),
            [last version that we checked at the time of writing](https://pint.readthedocs.io/en/0.24.4/user/plotting.html)).

        time_units
            Units of the time axis of the data.

            These are required if `unit_aware` is not `False`.

        x_label
            Label to apply to the x-axis.

            If `None`, no label will be applied.

        y_label
            Label to apply to the y-axis.

            If `True`, we will try and infer the y-label based on the data's units.

            If `None`, no label will be applied.

        warn_infer_y_label_with_multi_unit
            Should a warning be raised if we try to infer the y-unit
            but the data has more than one unit?

        create_legend
            Function to use to create the legend.

            This allows the user to have full control over the creation of the legend.

        observed
            Passed to [pd.DataFrame.groupby][pandas.DataFrame.groupby].

        Returns
        -------
        :
            Axes on which the data was plotted
        """
        return plot_plume_func(
            self._df,
            ax=ax,
            quantiles_plumes=quantiles_plumes,
            quantile_var=quantile_var,
            quantile_var_label=quantile_var_label,
            quantile_legend_round=quantile_legend_round,
            hue_var=hue_var,
            hue_var_label=hue_var_label,
            palette=palette,
            warn_on_palette_value_missing=warn_on_palette_value_missing,
            style_var=style_var,
            style_var_label=style_var_label,
            dashes=dashes,
            warn_on_dashes_value_missing=warn_on_dashes_value_missing,
            linewidth=linewidth,
            unit_var=unit_var,
            unit_aware=unit_aware,
            time_units=time_units,
            x_label=x_label,
            y_label=y_label,
            warn_infer_y_label_with_multi_unit=warn_infer_y_label_with_multi_unit,
            create_legend=create_legend,
            observed=observed,
        )

    def plot_plume_after_calculating_quantiles(  # noqa: PLR0913
        self,
        ax: matplotlib.axes.Axes | None = None,
        *,
        quantile_over: str | list[str],
        quantiles_plumes: QUANTILES_PLUMES_LIKE = (
            (0.5, 0.7),
            ((0.05, 0.95), 0.2),
        ),
        quantile_var_label: str | None = None,
        quantile_legend_round: int = 2,
        hue_var: str = "scenario",
        hue_var_label: str | None = None,
        palette: PALETTE_LIKE[Any] | None = None,
        warn_on_palette_value_missing: bool = True,
        style_var: str = "variable",
        style_var_label: str | None = None,
        dashes: dict[Any, str | tuple[float, tuple[float, ...]]] | None = None,
        warn_on_dashes_value_missing: bool = True,
        linewidth: float = 3.0,
        unit_var: str = "unit",
        unit_aware: bool | pint.facets.PlainRegistry = False,
        time_units: str | None = None,
        x_label: str | None = "time",
        y_label: str | bool | None = True,
        warn_infer_y_label_with_multi_unit: bool = True,
        create_legend: Callable[
            [matplotlib.axes.Axes, list[matplotlib.artist.Artist]], None
        ] = create_legend_default,
        observed: bool = True,
    ) -> matplotlib.axes.Axes:
        """
        Plot a plume plot, calculating the required quantiles first

        Parameters
        ----------
        ax
            Axes on which to plot.

            If not supplied, a new axes is created.

        quantile_over
            Variable(s)/column(s) over which to calculate the quantiles.

            The data is grouped by all columns except `quantile_over`
            when calculating the quantiles.

        quantiles_plumes
            Quantiles to plot in each plume.

            If the first element of each tuple is a tuple,
            a plume is plotted between the given quantiles.
            Otherwise, if the first element is a plain float,
            a line is plotted for the given quantile.

        quantile_var_label
            Label to use as the header for the quantile section in the legend

        quantile_legend_round
            Rounding to apply to quantile values when creating the legend

        hue_var
            Variable to use for grouping data into different colour groups

        hue_var_label
            Label to use as the header for the hue/colour section in the legend

        palette
            Colour to use for the different groups in the data.

            If any groups are not included in `palette`,
            they are auto-filled.

        warn_on_palette_value_missing
            Should a warning be emitted if there are values missing from `palette`?

        style_var
            Variable to use for grouping data into different (line)style groups

        style_var_label
            Label to use as the header for the style section in the legend

        dashes
            Dash/linestyle to use for the different groups in the data.

            If any groups are not included in `dashes`,
            they are auto-filled.

        warn_on_dashes_value_missing
            Should a warning be emitted if there are values missing from `dashes`?

        linewidth
            Width to use for plotting lines.

        unit_var
            Variable/column in the multi-index which stores information
            about the unit of each timeseries.

        unit_aware
            Should the plot be done in a unit-aware way?

            If `True`, we use the default application registry
            (retrieved with [pint.get_application_registry][]).
            Otherwise, a [pint.facets.PlainRegistry][] can be supplied and will be used.

            For details, see matplotlib and pint support plotting with units
            ([stable docs](https://pint.readthedocs.io/en/stable/user/plotting.html),
            [last version that we checked at the time of writing](https://pint.readthedocs.io/en/0.24.4/user/plotting.html)).

        time_units
            Units of the time axis.

            These are required if `unit_aware` is not `False`.

        x_label
            Label to apply to the x-axis.

            If `None`, no label will be applied.

        y_label
            Label to apply to the y-axis.

            If `True`, we will try and infer the y-label based on the data's units.

            If `None`, no label will be applied.

        warn_infer_y_label_with_multi_unit
            Should a warning be raised if we try to infer the y-unit
            but the data has more than one unit?

        create_legend
            Function to use to create the legend.

            This allows the user to have full control over the creation of the legend.

        observed
            Passed to [pd.DataFrame.groupby][pandas.DataFrame.groupby].

        Returns
        -------
        :
            Axes on which the data was plotted
        """
        return plot_plume_after_calculating_quantiles_func(
            self._df,
            ax=ax,
            quantile_over=quantile_over,
            quantiles_plumes=quantiles_plumes,
            quantile_var_label=quantile_var_label,
            quantile_legend_round=quantile_legend_round,
            hue_var=hue_var,
            hue_var_label=hue_var_label,
            palette=palette,
            warn_on_palette_value_missing=warn_on_palette_value_missing,
            style_var=style_var,
            style_var_label=style_var_label,
            dashes=dashes,
            warn_on_dashes_value_missing=warn_on_dashes_value_missing,
            linewidth=linewidth,
            unit_var=unit_var,
            unit_aware=unit_aware,
            time_units=time_units,
            x_label=x_label,
            y_label=y_label,
            warn_infer_y_label_with_multi_unit=warn_infer_y_label_with_multi_unit,
            create_legend=create_legend,
            observed=observed,
        )

    def set_index_levels(
        self,
        levels_to_set: dict[str, Any | Collection[Any]],
        copy: bool = True,
    ) -> pd.DataFrame:
        """
        Set the index levels

        Parameters
        ----------
        levels_to_set
            Mapping of level names to values to set

        copy
            Should the [pd.DataFrame][pandas.DataFrame] be copied before returning?

        Returns
        -------
        :
            [pd.DataFrame][pandas.DataFrame] with updates applied to its index
        """
        return set_index_levels_func(
            self._df,
            levels_to_set=levels_to_set,
            copy=copy,
        )

    def to_category_index(self) -> pd.DataFrame:
        """
        Convert the index's values to categories

        This can save a lot of memory and improve the speed of processing.
        However, it comes with some pitfalls.
        For a nice discussion of some of them,
        see [this article](https://towardsdatascience.com/staying-sane-while-adopting-pandas-categorical-datatypes-78dbd19dcd8a/).

        Returns
        -------
        :
            [pd.DataFrame][pandas.DataFrame] with all index columns
            converted to category type.
        """
        return convert_index_to_category_index(self._df)

    def to_long_data(self, time_col_name: str = "time") -> pd.DataFrame:
        """
        Convert to long data

        Here, long data means that each row contains a single value,
        alongside metadata associated with that value
        (for more details, see e.g.
        https://data.europa.eu/apps/data-visualisation-guide/wide-versus-long-data).

        Parameters
        ----------
        time_col_name
            Name of the time column in the output

        Returns
        -------
        :
            DataFrame in long-form

        Examples
        --------
        >>> import numpy as np
        >>> import pandas as pd
        >>>
        >>> from pandas_openscm.accessors import register_pandas_accessors
        >>>
        >>> register_pandas_accessors()
        >>>
        >>> df = pd.DataFrame(
        ...     [
        ...         [1.1, 0.8, 1.2],
        ...         [2.1, np.nan, 8.4],
        ...         [2.3, 3.2, 3.0],
        ...         [1.2, 2.8, np.nan],
        ...     ],
        ...     columns=[2010.0, 2015.0, 2025.0],
        ...     index=pd.MultiIndex.from_tuples(
        ...         [
        ...             ("sa", np.nan, "K"),
        ...             ("sb", "v1", None),
        ...             ("sa", "v2", "W"),
        ...             ("sb", "v2", "W"),
        ...         ],
        ...         names=["scenario", "variable", "unit"],
        ...     ),
        ... )
        >>>
        >>> # Start with wide data
        >>> df
                                2010.0  2015.0  2025.0
        scenario variable unit
        sa       NaN      K        1.1     0.8     1.2
        sb       v1       NaN      2.1     NaN     8.4
        sa       v2       W        2.3     3.2     3.0
        sb       v2       W        1.2     2.8     NaN
        >>>
        >>> # Convert to long data
        >>> df.openscm.to_long_data()
           scenario variable unit    time  value
        0        sa      NaN    K  2010.0    1.1
        1        sb       v1  NaN  2010.0    2.1
        2        sa       v2    W  2010.0    2.3
        3        sb       v2    W  2010.0    1.2
        4        sa      NaN    K  2015.0    0.8
        5        sb       v1  NaN  2015.0    NaN
        6        sa       v2    W  2015.0    3.2
        7        sb       v2    W  2015.0    2.8
        8        sa      NaN    K  2025.0    1.2
        9        sb       v1  NaN  2025.0    8.4
        10       sa       v2    W  2025.0    3.0
        11       sb       v2    W  2025.0    NaN
        >>>
        >>> # Specify a different time column name
        >>> df.openscm.to_long_data(time_col_name="year")
           scenario variable unit    year  value
        0        sa      NaN    K  2010.0    1.1
        1        sb       v1  NaN  2010.0    2.1
        2        sa       v2    W  2010.0    2.3
        3        sb       v2    W  2010.0    1.2
        4        sa      NaN    K  2015.0    0.8
        5        sb       v1  NaN  2015.0    NaN
        6        sa       v2    W  2015.0    3.2
        7        sb       v2    W  2015.0    2.8
        8        sa      NaN    K  2025.0    1.2
        9        sb       v1  NaN  2025.0    8.4
        10       sa       v2    W  2025.0    3.0
        11       sb       v2    W  2025.0    NaN
        >>>
        >>> # The result is just a pandas DataFrame,
        >>> # so you can do whatever operations you want
        >>> # on the result.
        >>> # A common one is probably dropping all rows with NaN
        >>> df.openscm.to_long_data(time_col_name="year").dropna()
           scenario variable unit    year  value
        2        sa       v2    W  2010.0    2.3
        3        sb       v2    W  2010.0    1.2
        6        sa       v2    W  2015.0    3.2
        7        sb       v2    W  2015.0    2.8
        10       sa       v2    W  2025.0    3.0
        >>>
        >>> # or just rows with NaN in particular columns
        >>> df.openscm.to_long_data(time_col_name="year").dropna(subset=["variable"])
           scenario variable unit    year  value
        1        sb       v1  NaN  2010.0    2.1
        2        sa       v2    W  2010.0    2.3
        3        sb       v2    W  2010.0    1.2
        5        sb       v1  NaN  2015.0    NaN
        6        sa       v2    W  2015.0    3.2
        7        sb       v2    W  2015.0    2.8
        9        sb       v1  NaN  2025.0    8.4
        10       sa       v2    W  2025.0    3.0
        11       sb       v2    W  2025.0    NaN
        """
        return ts_to_long_data(self._df, time_col_name=time_col_name)

    def update_index_levels(
        self,
        updates: dict[Any, Callable[[Any], Any]],
        copy: bool = True,
        remove_unused_levels: bool = True,
    ) -> pd.DataFrame:
        """
        Update the index levels

        Parameters
        ----------
        updates
            Updates to apply to the index levels

            Each key is the index level to which the updates will be applied.
            Each value is a function which updates the levels to their new values.

        copy
            Should the [pd.DataFrame][pandas.DataFrame] be copied before returning?

        remove_unused_levels
            Remove unused levels before applying the update

            Specifically, call
            [pd.MultiIndex.remove_unused_levels][pandas.MultiIndex.remove_unused_levels].

            This avoids trying to update levels that aren't being used.

        Returns
        -------
        :
            [pd.DataFrame][pandas.DataFrame] with updates applied to its index
        """
        return update_index_levels_func(
            self._df,
            updates=updates,
            copy=copy,
            remove_unused_levels=remove_unused_levels,
        )

    def update_index_levels_from_other(
        self,
        update_sources: dict[
            Any, tuple[Any, Callable[[Any], Any] | dict[Any, Any] | pd.Series[Any]]
        ],
        copy: bool = True,
        remove_unused_levels: bool = True,
    ) -> pd.DataFrame:
        """
        Update the index levels based on other index levels

        Parameters
        ----------
        update_sources
            Updates to apply to `df`'s index

            Each key is the level to which the updates will be applied
            (or the level that will be created if it doesn't already exist).

            Each value is a tuple of which the first element
            is the level to use to generate the values (the 'source level')
            and the second is mapper of the form used by
            [pd.Index.map][pandas.Index.map]
            which will be applied to the source level
            to update/create the level of interest.

        copy
            Should the [pd.DataFrame][pandas.DataFrame] be copied before returning?

        remove_unused_levels
            Remove unused levels before applying the update

            Specifically, call
            [pd.MultiIndex.remove_unused_levels][pandas.MultiIndex.remove_unused_levels].

            This avoids trying to update levels that aren't being used.

        Returns
        -------
        :
            [pd.DataFrame][pandas.DataFrame] with updates applied to its index
        """
        return update_index_levels_from_other_func(
            self._df,
            update_sources=update_sources,
            copy=copy,
            remove_unused_levels=remove_unused_levels,
        )



"""
Type hints that are used throughout
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Union

import numpy as np
import numpy.typing as npt

if TYPE_CHECKING:
    import pint
    from typing_extensions import TypeAlias

NP_FLOAT_OR_INT: TypeAlias = Union[np.floating[Any], np.integer[Any]]
"""
Type alias for a numpy float or int (not complex)
"""

NP_ARRAY_OF_FLOAT_OR_INT: TypeAlias = npt.NDArray[NP_FLOAT_OR_INT]
"""
Type alias for an array of numpy float or int (not complex)
"""

NP_ARRAY_OF_BOOL: TypeAlias = npt.NDArray[np.bool]
"""
Type alias for an array of boolean values
"""

if TYPE_CHECKING:
    PINT_NUMPY_ARRAY: TypeAlias = pint.facets.numpy.quantity.NumpyQuantity[
        NP_ARRAY_OF_FLOAT_OR_INT
    ]
    """
    Type alias for a pint quantity that wraps a numpy array

    No shape hints because that doesn't seem to be supported by numpy yet.
    """

    PINT_SCALAR: TypeAlias = pint.facets.numpy.quantity.NumpyQuantity[NP_FLOAT_OR_INT]
    """
    Type alias for a pint quantity that wraps a numpy scalar
    """



"""
Integration tests of `pandas_openscm.io`
"""

from __future__ import annotations

import random

import numpy as np
import pytest

from pandas_openscm.io import load_timeseries_csv
from pandas_openscm.testing import create_test_df


@pytest.mark.parametrize(
    "index_columns",
    (
        ["variable", "scenario", "run", "unit"],
        ["scenario", "variable", "unit", "run"],
        ["variable", "scenario", "run", "unit", "1990.0", "2005.0"],
        ["scenario", "run", "unit", "1990.0", "2005.0"],
    ),
)
def test_load_timeseries_csv_basic(tmp_path, index_columns):
    out_path = tmp_path / "test_load_timeseries_csv.csv"

    timepoints = np.arange(1990.0, 2010.0 + 1.0)
    start = create_test_df(
        variables=[(f"variable_{i}", "Mt") for i in range(5)],
        n_scenarios=3,
        n_runs=6,
        timepoints=timepoints,
    )

    # Write in a way that definitely doesn't preserve index information
    # or column type.
    start.reset_index().to_csv(out_path, index=False)

    loaded = load_timeseries_csv(out_path, index_columns=index_columns)

    assert loaded.index.names == index_columns
    # No mangling done
    assert all(isinstance(c, str) for c in loaded.columns.values)

    all_cols_as_str = [str(v) for v in [*start.index.names, *timepoints]]
    non_index_cols = set(all_cols_as_str) - set(index_columns)
    loaded_cols = loaded.columns.tolist()
    assert all(str(v) in loaded_cols for v in non_index_cols)


@pytest.mark.parametrize("lower_column_names", (True, False))
def test_load_timeseries_csv_lower_column_names(tmp_path, lower_column_names):
    out_path = tmp_path / "test_load_timeseries_csv_lower_column_names.csv"

    timepoints = np.arange(1990.0, 2010.0 + 1.0)
    start = create_test_df(
        variables=[(f"variable_{i}", "Mt") for i in range(5)],
        n_scenarios=3,
        n_runs=6,
        timepoints=timepoints,
    )

    # Write with capitalised columns
    to_write = start.reset_index()
    to_write.columns = [
        v.capitalize() if isinstance(v, str) else v for v in to_write.columns
    ]
    to_write.to_csv(out_path, index=False)

    if lower_column_names:
        index_columns = ["variable", "scenario", "run", "unit"]
    else:
        index_columns = ["Variable", "Scenario", "Run", "Unit"]

    loaded = load_timeseries_csv(
        out_path, index_columns=index_columns, lower_column_names=lower_column_names
    )

    assert loaded.index.names == index_columns
    assert all(isinstance(c, str) for c in loaded.columns.values)
    assert all(str(v) in loaded.columns for v in timepoints)


@pytest.mark.parametrize(
    # Column type and value type are not the same
    # because columns are held as numpy arrays.
    "out_columns_type, exp_column_value_type",
    (
        (int, np.int64),
        (float, np.float64),
        (np.float64, np.float64),
        (np.float32, np.float32),
    ),
)
def test_load_timeseries_csv_basic_out_columns_type(
    tmp_path, out_columns_type, exp_column_value_type
):
    out_path = tmp_path / "test_load_timeseries_csv.csv"

    timepoints = np.arange(1990.0, 2010.0 + 1.0, dtype=int)
    start = create_test_df(
        variables=[(f"variable_{i}", "Mt") for i in range(5)],
        n_scenarios=3,
        n_runs=6,
        timepoints=timepoints,
    )

    # Write in a way that definitely doesn't preserve index information
    # or column type.
    start.reset_index().to_csv(out_path, index=False)

    index_columns = ["variable", "scenario", "run", "unit"]

    loaded = load_timeseries_csv(
        out_path, index_columns=index_columns, out_columns_type=out_columns_type
    )

    assert loaded.index.names == index_columns
    assert all(isinstance(c, exp_column_value_type) for c in loaded.columns.values)


@pytest.mark.parametrize(
    "out_columns_name, exp_columns_name",
    (
        (None, None),
        ("hi", "hi"),
        ("time", "time"),
    ),
)
def test_load_timeseries_csv_basic_out_columns_name(
    tmp_path, out_columns_name, exp_columns_name
):
    out_path = tmp_path / "test_load_timeseries_csv.csv"

    timepoints = np.arange(1990.0, 2010.0 + 1.0, dtype=int)
    start = create_test_df(
        variables=[(f"variable_{i}", "Mt") for i in range(5)],
        n_scenarios=3,
        n_runs=6,
        timepoints=timepoints,
    )
    assert start.columns.name is None

    start.to_csv(out_path)

    index_columns = ["variable", "scenario", "run", "unit"]

    loaded = load_timeseries_csv(
        out_path, index_columns=index_columns, out_columns_name=out_columns_name
    )

    assert loaded.columns.name == exp_columns_name


@pytest.mark.xfail(reason="Not implemented")
def test_load_timeseries_csv_infer_index_cols(tmp_path):
    # Suggested cases here:
    # - datetime columns
    # - int columns
    # - float columns
    # - array columns
    out_path = tmp_path / "test_load_timeseries_csv.csv"

    timepoints = np.arange(1990.0, 2010.0 + 1.0)
    start = create_test_df(
        variables=[(f"variable_{i}", "Mt") for i in range(5)],
        n_scenarios=3,
        n_runs=6,
        timepoints=timepoints,
    )

    # Write in a way that definitely doesn't preserve index information
    # or column type and shuffles the columns order.
    to_write = start.reset_index()
    cols = to_write.columns.tolist()
    random.shuffle(cols)
    to_write = to_write[cols]
    to_write.to_csv(out_path, index=False)

    loaded = load_timeseries_csv(out_path)

    exp_index_columns = ["scenario", "variable", "unit", "run"]
    assert loaded.index.names == exp_index_columns
    assert all(isinstance(c, str) for c in loaded.columns.values)
    assert all(str(v) in loaded.columns for v in timepoints)



"""
Available back-ends

This is just a shortcut/convenience module
"""

from __future__ import annotations

from pathlib import Path

from attrs import frozen

from pandas_openscm.db.csv import CSVDataBackend, CSVIndexBackend
from pandas_openscm.db.feather import FeatherDataBackend, FeatherIndexBackend
from pandas_openscm.db.in_memory import InMemoryDataBackend, InMemoryIndexBackend
from pandas_openscm.db.interfaces import OpenSCMDBDataBackend, OpenSCMDBIndexBackend
from pandas_openscm.db.netcdf import netCDFDataBackend, netCDFIndexBackend


@frozen
class DataBackendOptions:
    """A collection of data back-end options"""

    options: tuple[  # type hint doesn't work properly, but ok
        tuple[str, type[OpenSCMDBDataBackend]], ...
    ]
    """
    Options

    The first element of each option is the option's short name.
    The second element is the class that matches that option.
    """

    def get_instance(self, option: str) -> OpenSCMDBDataBackend:
        """
        Get an instance of one of the options

        Parameters
        ----------
        option
            Option for which to get a data back-end instance

        Returns
        -------
        :
            Initialised instance

        Raises
        ------
        KeyError
            The option is not supported
        """
        for short_name, option_cls in self.options:
            if short_name == option:
                return option_cls()

        msg = (
            f"{option=} is not supported. "
            f"Available options: {tuple(v[1] for v in self.options)}"
        )
        raise KeyError(msg)

    def guess_backend(self, data_file_name: str) -> OpenSCMDBDataBackend:
        """
        Guess backend from a file name

        Parameters
        ----------
        data_file_name
            Name of the data file from which to guess the backend

        Returns
        -------
        :
            Guessed backend

        Raises
        ------
        ValueError
            The backend could not be guessed from `data_file_name`
        """
        ext = Path(data_file_name).suffix
        for _, option_cls in self.options:
            option = option_cls()
            if ext == option.ext:
                return option

        known_options_and_extensions = [(v[0], v[1]().ext) for v in self.options]
        msg = (
            f"Could not guess backend from {data_file_name=!r}. "
            "The file's extension does not match any of the available options: "
            f"{known_options_and_extensions=}"
        )
        raise ValueError(msg)


DATA_BACKENDS = DataBackendOptions(
    (  # type: ignore # using class with protocol doesn't work properly
        ("csv", CSVDataBackend),
        ("feather", FeatherDataBackend),
        ("in_memory", InMemoryDataBackend),
        ("netCDF", netCDFDataBackend),
        # Other options to consider:
        #
        # - pretty netCDF, where we try and save the data with dimensions where possible
        #
        # - HDF5: https://pandas.pydata.org/docs/user_guide/io.html#hdf5-pytables
        # - sqllite
    )
)
"""Inbuilt data back-ends"""


@frozen
class IndexBackendOptions:
    """A collection of index back-end options"""

    options: tuple[tuple[str, type[OpenSCMDBIndexBackend]], ...]
    """
    Options

    The first element of each option is the option's short name.
    The second element is the class that matches that option.
    """

    def get_instance(self, option: str) -> OpenSCMDBIndexBackend:
        """
        Get an instance of one of the options

        Parameters
        ----------
        option
            Option for which to get a index back-end instance

        Returns
        -------
        :
            Initialised instance

        Raises
        ------
        KeyError
            The option is not supported
        """
        for short_name, option_cls in self.options:
            if short_name == option:
                return option_cls()

        msg = (
            f"{option=} is not supported. "
            f"Available options: {tuple(v[1] for v in self.options)}"
        )
        raise KeyError(msg)

    def guess_backend(self, index_file_name: str) -> OpenSCMDBIndexBackend:
        """
        Guess backend from a file name

        Parameters
        ----------
        index_file_name
            Name of the index file from which to guess the backend

        Returns
        -------
        :
            Guessed backend

        Raises
        ------
        ValueError
            The backend could not be guessed from `index_file_name`
        """
        ext = Path(index_file_name).suffix
        for _, option_cls in self.options:
            option = option_cls()
            if ext == option.ext:
                return option

        known_options_and_extensions = [(v[0], v[1]().ext) for v in self.options]
        msg = (
            f"Could not guess backend from {index_file_name=!r}. "
            "The file's extension does not match any of the available options: "
            f"{known_options_and_extensions=}"
        )
        raise ValueError(msg)


INDEX_BACKENDS = IndexBackendOptions(
    (  # type: ignore # using class with protocol doesn't work properly
        ("csv", CSVIndexBackend),
        ("feather", FeatherIndexBackend),
        ("in_memory", InMemoryIndexBackend),
        ("netCDF", netCDFIndexBackend),
        # Other options to consider:
        #
        # - HDF5: https://pandas.pydata.org/docs/user_guide/io.html#hdf5-pytables
        # - sqllite
    )
)
"""Inbuilt index back-ends"""



"""
Tests of `pandas_openscm.comparison`
"""

from functools import partial

import numpy as np
import pandas as pd
import pytest

from pandas_openscm.comparison import compare_close


@pytest.fixture
def simple_df():
    return pd.DataFrame(
        data=[[1.0, 2.0, 3.0], [1.1, 1.2, 1.3], [-1.1, 0.0, 0.5]],
        columns=pd.Index([2.0, 4.0, 10.0], name="time"),
        index=pd.MultiIndex.from_tuples(
            [("v1", "kg"), ("v2", "m"), ("v3", "yr")], names=["variable", "unit"]
        ),
    )


def test_equal(simple_df):
    assert compare_close(
        simple_df, simple_df, left_name="left", right_name="right"
    ).empty


def test_equal_misaligned(simple_df):
    left = simple_df

    # switch rows 2 and 3
    right = left.iloc[[0, 2, 1], :].copy()

    assert compare_close(left, right, left_name="left", right_name="right").empty


def create_test_df(
    data,
    columns=(2.0, 4.0, 10.0),
    index=(("v1", "kg"), ("v2", "m"), ("v3", "yr")),
):
    return pd.DataFrame(
        data=data,
        columns=pd.Index(columns, name="time"),
        index=pd.MultiIndex.from_tuples(index, names=["variable", "unit"]),
    )


@pytest.mark.parametrize(
    "left, right, left_name, right_name, isclose, exp",
    (
        pytest.param(
            create_test_df(
                data=[[1.0, 2.0, 3.0], [1.1, 1.2, 1.3], [-1.1, 0.0, 0.5]],
            ),
            create_test_df(
                data=[[1.5, 2.0, 3.0], [1.1, 1.2, 1.3], [-1.1, 0.0, 0.5]],
            ),
            "left",
            "right",
            partial(np.isclose, rtol=0.1),
            pd.DataFrame(
                data=[[1.0, 1.5]],
                columns=["left", "right"],
                index=pd.MultiIndex.from_tuples(
                    [("v1", "kg", 2.0)], names=["variable", "unit", "time"]
                ),
            ),
            id="values outside relative tolerance",
        ),
        pytest.param(
            create_test_df(
                data=[[1.0, 2.0, 3.0], [1.1, 1.2, 1.3], [-1.1, 0.0, 0.5]],
            ),
            create_test_df(
                data=[[1.5, 2.0, 3.0], [1.1, 1.2, 1.3], [-1.1, 0.0, 0.5]],
            ),
            "name1_left",
            "name2_right",
            partial(np.isclose, rtol=0.6),
            pd.DataFrame(
                index=pd.MultiIndex.from_tuples([], names=["variable", "unit", "time"]),
                columns=["name1_left", "name2_right"],
            ),
            id="values within relative tolerance",
        ),
        pytest.param(
            create_test_df(
                data=[[1.0, 2.0, 3.0], [1.1, 1.2, 1.7], [-1.1, 0.0, 0.5]],
            ),
            create_test_df(
                data=[[1.5, 2.0, 3.5], [1.1, 1.2, 1.3], [-1.1, 0.0, 0.5]],
            ),
            "name1_left",
            "name2_right",
            partial(np.isclose, atol=0.5),
            pd.DataFrame(
                index=pd.MultiIndex.from_tuples([], names=["variable", "unit", "time"]),
                columns=["name1_left", "name2_right"],
            ),
            id="values within absolute tolerance - edge case",
        ),
    ),
)
def test_compare_close(left, right, left_name, right_name, isclose, exp):  # noqa: PLR0913
    pd.testing.assert_frame_equal(
        compare_close(left, right, left_name, right_name, isclose),
        exp,
        #  when comparing empty dataframes, index type and dtype may be different
        check_index_type=False,
        check_dtype=False,
    )



"""
Manipulation of the index of data
"""

from __future__ import annotations

from collections.abc import Collection, Mapping
from typing import TYPE_CHECKING, Any, Callable, TypeVar

import numpy as np
import numpy.typing as npt
import pandas as pd

if TYPE_CHECKING:
    P = TypeVar("P", pd.DataFrame, pd.Series[Any])

    import pandas.core.indexes.frozen


def convert_index_to_category_index(pandas_obj: P) -> P:
    """
    Convert the index's values to categories

    This can save a lot of memory and improve the speed of processing.
    However, it comes with some pitfalls.
    For a nice discussion of some of them,
    see [this article](https://towardsdatascience.com/staying-sane-while-adopting-pandas-categorical-datatypes-78dbd19dcd8a/).

    Parameters
    ----------
    pandas_obj
        Object whose index we want to change to categorical.

    Returns
    -------
    :
        A new object with the same data as `pandas_obj`
        but a category type index.
    """
    new_index = pd.MultiIndex.from_frame(
        pandas_obj.index.to_frame(index=False).astype("category")
    )

    if hasattr(pandas_obj, "columns"):
        return type(pandas_obj)(  # type: ignore # confusing mypy here
            pandas_obj.values,
            index=new_index,
            columns=pandas_obj.columns,
        )

    return type(pandas_obj)(
        pandas_obj.values,
        index=new_index,
    )


def ensure_is_multiindex(index: pd.Index[Any] | pd.MultiIndex) -> pd.MultiIndex:
    """
    Ensure that an index is a [pd.MultiIndex][pandas.MultiIndex]

    Parameters
    ----------
    index
        Index to check

    Returns
    -------
    :
        Index, cast to [pd.MultiIndex][pandas.MultiIndex] if needed
    """
    if isinstance(index, pd.MultiIndex):
        return index

    return pd.MultiIndex.from_arrays([index.values], names=[index.name])


def ensure_index_is_multiindex(pandas_obj: P, copy: bool = True) -> P:
    """
    Ensure that the index of a pandas object is a [pd.MultiIndex][pandas.MultiIndex]

    Parameters
    ----------
    pandas_obj
        Object whose index we want to ensure is a [pd.MultiIndex][pandas.MultiIndex]

    copy
        Should we copy `pandas_obj` before modifying the index?

    Returns
    -------
    :
        `pandas_obj` with a [pd.MultiIndex][pandas.MultiIndex]

        If the index was already a [pd.MultiIndex][pandas.MultiIndex],
        this is a no-op (although the value of copy is respected).
    """
    if copy:
        pandas_obj = pandas_obj.copy()

    if isinstance(pandas_obj.index, pd.MultiIndex):
        return pandas_obj

    pandas_obj.index = ensure_is_multiindex(pandas_obj.index)

    return pandas_obj


def unify_index_levels(
    left: pd.MultiIndex, right: pd.MultiIndex
) -> tuple[pd.MultiIndex, pd.MultiIndex]:
    """
    Unify the levels on two indexes

    The levels are unified by simply adding NaN to any level in either `left` or `right`
    that is not in the level of the other index.

    This is differnt to [pd.DataFrame.align][pandas.DataFrame.align].
    [pd.DataFrame.align][pandas.DataFrame.align]
    will fill missing values with values from the other index if it can.
    We don't want that here.
    We want any non-aligned levels to be filled with NaN.

    The implementation also allows this to be performed on indexes directly
    (avoiding casting to a DataFrame
    and avoiding paying the price of aligning everything else
    or creating a bunch of NaN that we just drop straight away).

    The indexes are returned with the levels from `left` first,
    then the levels from `right`.

    Parameters
    ----------
    left
        First index to unify

    right
        Second index to unify

    Returns
    -------
    left_aligned :
        Left after alignment

    right_aligned :
        Right after alignment

    Examples
    --------
    >>> import pandas as pd
    >>>
    >>> idx_a = pd.MultiIndex.from_tuples(
    ...     [
    ...         (1, 2, 3),
    ...         (4, 5, 6),
    ...     ],
    ...     names=["a", "b", "c"],
    ... )
    >>> idx_b = pd.MultiIndex.from_tuples(
    ...     [
    ...         (7, 8),
    ...         (10, 11),
    ...     ],
    ...     names=["a", "b"],
    ... )
    >>> unified_a, unified_b = unify_index_levels(idx_a, idx_b)
    >>> unified_a
    MultiIndex([(1, 2, 3),
                (4, 5, 6)],
               names=['a', 'b', 'c'])

    >>> unified_b
    MultiIndex([( 7,  8, nan),
                (10, 11, nan)],
               names=['a', 'b', 'c'])

    >>> # Also fine if b has swapped levels
    >>> idx_b = pd.MultiIndex.from_tuples(
    ...     [
    ...         (7, 8),
    ...         (10, 11),
    ...     ],
    ...     names=["b", "a"],
    ... )
    >>> unified_a, unified_b = unify_index_levels(idx_a, idx_b)
    >>> unified_a
    MultiIndex([(1, 2, 3),
                (4, 5, 6)],
               names=['a', 'b', 'c'])

    >>> unified_b
    MultiIndex([( 8,  7, nan),
                (11, 10, nan)],
               names=['a', 'b', 'c'])

    >>> # Also works if a is 'inside' b
    >>> idx_a = pd.MultiIndex.from_tuples(
    ...     [
    ...         (7, 8),
    ...         (10, 11),
    ...     ],
    ...     names=["a", "b"],
    ... )
    >>> idx_b = pd.MultiIndex.from_tuples(
    ...     [
    ...         (1, 2, 3),
    ...         (4, 5, 6),
    ...     ],
    ...     names=["a", "b", "c"],
    ... )
    >>> unified_a, unified_b = unify_index_levels(idx_a, idx_b)
    >>> unified_a
    MultiIndex([( 7,  8, nan),
                (10, 11, nan)],
               names=['a', 'b', 'c'])

    >>> unified_b
    MultiIndex([(1, 2, 3),
                (4, 5, 6)],
               names=['a', 'b', 'c'])

    >>> # But, be a bit careful, this is now sensitive to a's column order
    >>> idx_a = pd.MultiIndex.from_tuples(
    ...     [
    ...         (7, 8),
    ...         (10, 11),
    ...     ],
    ...     names=["b", "a"],
    ... )
    >>> idx_b = pd.MultiIndex.from_tuples(
    ...     [
    ...         (1, 2, 3),
    ...         (4, 5, 6),
    ...     ],
    ...     names=["a", "b", "c"],
    ... )
    >>> unified_a, unified_b = unify_index_levels(idx_a, idx_b)
    >>> # Note that the names are `['b', 'a', 'c']` in the output
    >>> unified_a
    MultiIndex([( 7,  8, nan),
                (10, 11, nan)],
               names=['b', 'a', 'c'])

    >>> unified_b
    MultiIndex([(2, 1, 3),
                (5, 4, 6)],
               names=['b', 'a', 'c'])
    """
    if left.names == right.names:
        return left, right

    if (not left.names.difference(right.names)) and (  # type: ignore # pandas-stubs confused
        not right.names.difference(left.names)  # type: ignore # pandas-stubs confused
    ):
        return left, right.reorder_levels(left.names)  # type: ignore # pandas-stubs missing reorder_levels

    out_names = [*left.names, *[v for v in right.names if v not in left.names]]
    out_names_s = set(out_names)
    left_to_add = out_names_s.difference(left.names)
    right_to_add = out_names_s.difference(right.names)

    left_unified = pd.MultiIndex(  # type: ignore # pandas-stubs missing reorder_levels
        levels=[
            *left.levels,
            *[np.array([], dtype=right.get_level_values(c).dtype) for c in left_to_add],  # type: ignore # pandas-stubs confused
        ],
        codes=[
            *left.codes,
            *([np.full(left.shape[0], -1)] * len(left_to_add)),
        ],
        names=[
            *left.names,
            *left_to_add,
        ],
    ).reorder_levels(out_names)

    right_unified = pd.MultiIndex(  # type: ignore # pandas-stubs missing reorder_levels
        levels=[
            *[np.array([], dtype=left.get_level_values(c).dtype) for c in right_to_add],  # type: ignore # pandas-stubs confused
            *right.levels,
        ],
        codes=[
            *([np.full(right.shape[0], -1)] * len(right_to_add)),
            *right.codes,
        ],
        names=[
            *right_to_add,
            *right.names,
        ],
    ).reorder_levels(out_names)

    return left_unified, right_unified


def unify_index_levels_check_index_types(
    left: pd.Index[Any], right: pd.Index[Any]
) -> tuple[pd.MultiIndex, pd.MultiIndex]:
    """
    Unify the levels on two indexes

    This is just a thin wrapper around [unify_index_levels][(m).]
    that checks the the inputs are both [pd.MultiIndex][pandas.MultiIndex]
    before unifying the indices.

    Parameters
    ----------
    left
        First index to unify

    right
        Second index to unify

    Returns
    -------
    left_aligned :
        Left after alignment

    right_aligned :
        Right after alignment
    """
    if not isinstance(left, pd.MultiIndex):
        raise TypeError(left)

    if not isinstance(right, pd.MultiIndex):
        raise TypeError(right)

    return unify_index_levels(left, right)


def update_index_from_candidates(
    indf: pd.DataFrame, candidates: pandas.core.indexes.frozen.FrozenList
) -> pd.DataFrame:
    """
    Update the index of data to align with the candidate columns as much as possible

    Parameters
    ----------
    indf
        Data of which to update the index

    candidates
        Candidate columns to use to create the updated index

    Returns
    -------
    :
        `indf` with its updated index.

        All columns of `indf` that are in `candidates`
        are used to create the index of the result.

    Notes
    -----
    This overwrites any existing index of `indf`
    so you will only want to use this function
    when you're sure that there isn't anything of interest
    already in the index of `indf`.
    """
    set_to_index = [v for v in candidates if v in indf.columns]
    res = indf.set_index(set_to_index)

    return res


def create_new_level_and_codes_by_mapping(
    ini: pd.MultiIndex,
    level_to_create_from: str,
    mapper: Callable[[Any], Any] | dict[Any, Any] | pd.Series[Any],
) -> tuple[pd.Index[Any], npt.NDArray[np.integer[Any]]]:
    """
    Create a new level and associated codes by mapping an existing level

    This is a thin function intended for internal use
    to handle some slightly tricky logic.

    Parameters
    ----------
    ini
        Input index

    level_to_create_from
        Level to create the new level from

    mapper
        Function to use to map existing levels to new levels

    Returns
    -------
    new_level :
        New level

    new_codes :
        New codes
    """
    level_to_map_from_idx = ini.names.index(level_to_create_from)
    new_level = ini.levels[level_to_map_from_idx].map(mapper)
    if not new_level.has_duplicates:
        # Fast route, can just return new level and codes from level we mapped from
        return new_level, ini.codes[level_to_map_from_idx]

    # Slow route: have to update the codes
    dup_level = ini.get_level_values(level_to_create_from).map(mapper)
    new_level = new_level.unique()
    new_codes = new_level.get_indexer(dup_level)  # type: ignore

    return new_level, new_codes


def update_index_levels_func(
    df: pd.DataFrame,
    updates: Mapping[Any, Callable[[Any], Any] | dict[Any, Any] | pd.Series[Any]],
    copy: bool = True,
    remove_unused_levels: bool = True,
) -> pd.DataFrame:
    """
    Update the index levels of a [pd.DataFrame][pandas.DataFrame]

    Parameters
    ----------
    df
        [pd.DataFrame][pandas.DataFrame] to update

    updates
        Updates to apply to `df`'s index

        Each key is the index level to which the updates will be applied.
        Each value is a function which updates the levels to their new values.

    copy
        Should `df` be copied before returning?

    remove_unused_levels
        Call `df.index.remove_unused_levels` before updating the levels

        This avoids trying to update levels that aren't being used.

    Returns
    -------
    :
        `df` with updates applied to its index
    """
    if copy:
        df = df.copy()

    if not isinstance(df.index, pd.MultiIndex):
        msg = (
            "This function is only intended to be used "
            "when `df`'s index is an instance of `MultiIndex`. "
            f"Received {type(df.index)=}"
        )
        raise TypeError(msg)

    df.index = update_levels(
        df.index, updates=updates, remove_unused_levels=remove_unused_levels
    )

    return df


def update_levels(
    ini: pd.MultiIndex,
    updates: Mapping[Any, Callable[[Any], Any] | dict[Any, Any] | pd.Series[Any]],
    remove_unused_levels: bool = True,
) -> pd.MultiIndex:
    """
    Update the levels of a [pd.MultiIndex][pandas.MultiIndex]

    Parameters
    ----------
    ini
        Input index

    updates
        Updates to apply

        Each key is the level to which the updates will be applied.
        Each value is a mapper of the form used by
        [pd.Index.map][pandas.Index.map].

    remove_unused_levels
        Call `ini.remove_unused_levels` before updating the levels

        This avoids trying to update levels that aren't being used.

    Returns
    -------
    :
        `ini` with updates applied

    Raises
    ------
    KeyError
        A level in `updates` is not a level in `ini`

    Examples
    --------
    >>> start = pd.MultiIndex.from_tuples(
    ...     [
    ...         ("sa", "ma", "v1", "kg"),
    ...         ("sb", "ma", "v2", "m"),
    ...         ("sa", "mb", "v1", "kg"),
    ...         ("sa", "mb", "v2", "m"),
    ...     ],
    ...     names=["scenario", "model", "variable", "unit"],
    ... )
    >>> start
    MultiIndex([('sa', 'ma', 'v1', 'kg'),
                ('sb', 'ma', 'v2',  'm'),
                ('sa', 'mb', 'v1', 'kg'),
                ('sa', 'mb', 'v2',  'm')],
               names=['scenario', 'model', 'variable', 'unit'])
    >>>
    >>> update_levels(
    ...     start,
    ...     {"model": lambda x: f"model {x}", "scenario": lambda x: f"scenario {x}"},
    ... )
    MultiIndex([('scenario sa', 'model ma', 'v1', 'kg'),
                ('scenario sb', 'model ma', 'v2',  'm'),
                ('scenario sa', 'model mb', 'v1', 'kg'),
                ('scenario sa', 'model mb', 'v2',  'm')],
               names=['scenario', 'model', 'variable', 'unit'])
    >>>
    >>> update_levels(
    ...     start,
    ...     {"variable": {"v1": "variable one", "v2": "variable two"}},
    ... )
    MultiIndex([('sa', 'ma', 'variable one', 'kg'),
                ('sb', 'ma', 'variable two',  'm'),
                ('sa', 'mb', 'variable one', 'kg'),
                ('sa', 'mb', 'variable two',  'm')],
               names=['scenario', 'model', 'variable', 'unit'])
    """
    if remove_unused_levels:
        ini = ini.remove_unused_levels()  # type: ignore

    levels: list[pd.Index[Any]] = list(ini.levels)
    codes: list[npt.NDArray[np.integer[Any]]] = list(ini.codes)

    for level, updater in updates.items():
        if level not in ini.names:
            msg = (
                f"{level} is not available in the index. Available levels: {ini.names}"
            )
            raise KeyError(msg)

        new_level, new_codes = create_new_level_and_codes_by_mapping(
            ini=ini,
            level_to_create_from=level,
            mapper=updater,
        )

        level_idx = ini.names.index(level)
        levels[level_idx] = new_level
        codes[level_idx] = new_codes

    res = pd.MultiIndex(
        levels=levels,
        codes=codes,
        names=ini.names,
    )

    return res


def update_index_levels_from_other_func(
    df: pd.DataFrame,
    update_sources: dict[
        Any, tuple[Any, Callable[[Any], Any] | dict[Any, Any] | pd.Series[Any]]
    ],
    copy: bool = True,
    remove_unused_levels: bool = True,
) -> pd.DataFrame:
    """
    Update the index levels based on other levels of a [pd.DataFrame][pandas.DataFrame]

    If the level to be updated doesn't exist,
    it is created.

    Parameters
    ----------
    df
        [pd.DataFrame][pandas.DataFrame] to update

    update_sources
        Updates to apply to `df`'s index

        Each key is the level to which the updates will be applied
        (or the level that will be created if it doesn't already exist).

        Each value is a tuple of which the first element
        is the level to use to generate the values (the 'source level')
        and the second is mapper of the form used by
        [pd.Index.map][pandas.Index.map]
        which will be applied to the source level
        to update/create the level of interest.

    copy
        Should `df` be copied before returning?

    remove_unused_levels
        Call `df.index.remove_unused_levels` before updating the levels

        This avoids trying to update levels that aren't being used.

    Returns
    -------
    :
        `df` with updates applied to its index
    """
    if copy:
        df = df.copy()

    if not isinstance(df.index, pd.MultiIndex):
        msg = (
            "This function is only intended to be used "
            "when `df`'s index is an instance of `MultiIndex`. "
            f"Received {type(df.index)=}"
        )
        raise TypeError(msg)

    df.index = update_levels_from_other(
        df.index,
        update_sources=update_sources,
        remove_unused_levels=remove_unused_levels,
    )

    return df


def update_levels_from_other(
    ini: pd.MultiIndex,
    update_sources: dict[
        Any, tuple[Any, Callable[[Any], Any] | dict[Any, Any] | pd.Series[Any]]
    ],
    remove_unused_levels: bool = True,
) -> pd.MultiIndex:
    """
    Update levels based on other levels in a [pd.MultiIndex][pandas.MultiIndex]

    If the level to be updated doesn't exist,
    it is created.

    Parameters
    ----------
    ini
        Input index

    update_sources
        Updates to apply and their source levels

        Each key is the level to which the updates will be applied
        (or the level that will be created if it doesn't already exist).

        Each value is a tuple of which the first element
        is the level to use to generate the values (the 'source level')
        and the second is mapper of the form used by
        [pd.Index.map][pandas.Index.map]
        which will be applied to the source level
        to update/create the level of interest.

    remove_unused_levels
        Call `ini.remove_unused_levels` before updating the levels

        This avoids trying to update bsaed on levels that aren't being used.

    Returns
    -------
    :
        `ini` with updates applied

    Raises
    ------
    KeyError
        A source level in `update_sources` is not a level in `ini`

    Examples
    --------
    >>> start = pd.MultiIndex.from_tuples(
    ...     [
    ...         ("sa", "ma", "v1", "kg"),
    ...         ("sb", "ma", "v2", "m"),
    ...         ("sa", "mb", "v1", "kg"),
    ...         ("sa", "mb", "v2", "m"),
    ...     ],
    ...     names=["scenario", "model", "variable", "unit"],
    ... )
    >>> start
    MultiIndex([('sa', 'ma', 'v1', 'kg'),
                ('sb', 'ma', 'v2',  'm'),
                ('sa', 'mb', 'v1', 'kg'),
                ('sa', 'mb', 'v2',  'm')],
               names=['scenario', 'model', 'variable', 'unit'])
    >>>
    >>> # Create a new level based on an existing level
    >>> update_levels_from_other(
    ...     start,
    ...     {
    ...         "unit squared": ("unit", lambda x: f"{x}**2"),
    ...         "class": ("model", {"ma": "delta", "mb": "gamma"}),
    ...     },
    ... )
    MultiIndex([('sa', 'ma', 'v1', 'kg', 'kg**2', 'delta'),
                ('sb', 'ma', 'v2',  'm',  'm**2', 'delta'),
                ('sa', 'mb', 'v1', 'kg', 'kg**2', 'gamma'),
                ('sa', 'mb', 'v2',  'm',  'm**2', 'gamma')],
               names=['scenario', 'model', 'variable', 'unit', 'unit squared', 'class'])
    >>>
    >>> # Update an existing level based on another level
    >>> update_levels_from_other(
    ...     start,
    ...     {
    ...         "unit": ("variable", {"v1": "g", "v2": "km"}),
    ...         "model": ("scenario", lambda x: f"model {x}"),
    ...     },
    ... )
    MultiIndex([('sa', 'model sa', 'v1',  'g'),
                ('sb', 'model sb', 'v2', 'km'),
                ('sa', 'model sa', 'v1',  'g'),
                ('sa', 'model sa', 'v2', 'km')],
               names=['scenario', 'model', 'variable', 'unit'])
    >>>
    >>> # Both at the same time
    >>> update_levels_from_other(
    ...     start,
    ...     {
    ...         "title": ("scenario", lambda x: x.capitalize()),
    ...         "unit": ("unit", {"v1": "g", "v2": "km"}),
    ...     },
    ... )
    MultiIndex([('sa', 'ma', 'v1', nan, 'Sa'),
                ('sb', 'ma', 'v2', nan, 'Sb'),
                ('sa', 'mb', 'v1', nan, 'Sa'),
                ('sa', 'mb', 'v2', nan, 'Sa')],
               names=['scenario', 'model', 'variable', 'unit', 'title'])
    """
    if remove_unused_levels:
        ini = ini.remove_unused_levels()  # type: ignore

    levels: list[pd.Index[Any]] = list(ini.levels)
    codes: list[npt.NDArray[np.integer[Any]]] = list(ini.codes)
    names: list[str] = list(ini.names)

    for level, (source, updater) in update_sources.items():
        if source not in ini.names:
            msg = (
                f"{source} is not available in the index. Available levels: {ini.names}"
            )
            raise KeyError(msg)

        new_level, new_codes = create_new_level_and_codes_by_mapping(
            ini=ini,
            level_to_create_from=source,
            mapper=updater,
        )

        if level in ini.names:
            level_idx = ini.names.index(level)
            levels[level_idx] = new_level
            codes[level_idx] = new_codes

        else:
            levels.append(new_level)
            codes.append(new_codes)
            names.append(level)

    res = pd.MultiIndex(levels=levels, codes=codes, names=names)

    return res


def create_level_from_collection(
    level: str, value: Collection[Any]
) -> tuple[pandas.Index[Any], npt.NDArray[np.integer[Any]]]:
    """
    Create new level and corresponding codes.

    Parameters
    ----------
    level
        Name of the level to create

    value
        Values to use to create the level

    Returns
    -------
    :
        New level and corresponding codes
    """
    new_level: pandas.Index[Any] = pd.Index(value, name=level)
    if not new_level.has_duplicates:
        # Fast route, can just return new level and codes from level we mapped from
        return new_level, np.arange(len(value))

    # Slow route, have to update the codes
    new_level = new_level.unique()
    new_codes = new_level.get_indexer(value)  # type: ignore

    return new_level, new_codes


def set_levels(
    ini: pd.MultiIndex, levels_to_set: dict[str, Any | Collection[Any]]
) -> pd.MultiIndex:
    """
    Set the levels of a MultiIndex to the provided values

    Parameters
    ----------
    ini
        Input MultiIndex

    levels_to_set
        Mapping of level names to values to set. If values is of type `Collection`,
        it must be of the same length as the MultiIndex. If it is not a `Collection`,
        it will be set to the same value for all levels.

    Returns
    -------
    :
        New MultiIndex with the levels set to the provided values

    Raises
    ------
    TypeError
        If `ini` is not a MultiIndex
    ValueError
        If the length of the values is a collection that is not equal to the
        length of the index

    Examples
    --------
    >>> start = pd.MultiIndex.from_tuples(
    ...     [
    ...         ("sa", "ma", "v1", "kg"),
    ...         ("sb", "ma", "v2", "m"),
    ...         ("sa", "mb", "v1", "kg"),
    ...         ("sa", "mb", "v2", "m"),
    ...     ],
    ...     names=["scenario", "model", "variable", "unit"],
    ... )
    >>> start
    MultiIndex([('sa', 'ma', 'v1', 'kg'),
                ('sb', 'ma', 'v2',  'm'),
                ('sa', 'mb', 'v1', 'kg'),
                ('sa', 'mb', 'v2',  'm')],
               names=['scenario', 'model', 'variable', 'unit'])
    >>>
    >>> # Set a new level with a single string
    >>> set_levels(
    ...     start,
    ...     {"new_variable": "xyz"},
    ... )
    MultiIndex([('sa', 'ma', 'v1', 'kg', 'xyz'),
            ('sb', 'ma', 'v2',  'm', 'xyz'),
            ('sa', 'mb', 'v1', 'kg', 'xyz'),
            ('sa', 'mb', 'v2',  'm', 'xyz')],
           names=['scenario', 'model', 'variable', 'unit', 'new_variable'])
    >>>
    >>> # Replace a level with a collection
    >>> set_levels(
    ...     start,
    ...     {"new_variable": [1, 2, 3, 4]},
    ... )
    MultiIndex([('sa', 'ma', 'v1', 'kg', 1),
                ('sb', 'ma', 'v2',  'm', 2),
                ('sa', 'mb', 'v1', 'kg', 3),
                ('sa', 'mb', 'v2',  'm', 4)],
               names=['scenario', 'model', 'variable', 'unit', 'new_variable'])
    >>>
    >>> # Replace a level with a single value and add a new level
    >>> set_levels(
    ...     start,
    ...     {"model": "new_model", "new_variable": ["xyz", "xyz", "x", "y"]},
    ... )
    MultiIndex([('sa', 'new_model', 'v1', 'kg', 'xyz'),
                ('sb', 'new_model', 'v2',  'm', 'xyz'),
                ('sa', 'new_model', 'v1', 'kg',   'x'),
                ('sa', 'new_model', 'v2',  'm',   'y')],
               names=['scenario', 'model', 'variable', 'unit', 'new_variable'])
    """
    levels: list[pd.Index[Any]] = list(ini.levels)
    codes: list[npt.NDArray[np.integer[Any]]] = list(ini.codes)
    names: list[str] = list(ini.names)

    for level, value in levels_to_set.items():
        if isinstance(value, Collection) and not isinstance(value, str):
            if len(value) != len(ini):
                msg = (
                    f"Length of values for level '{level}' does not "
                    f"match index length: {len(value)} != {len(ini)}"
                )
                raise ValueError(msg)
            new_level, new_codes = create_level_from_collection(level, value)
        else:
            new_level = pd.Index([value], name=level)
            new_codes = np.zeros(ini.shape[0], dtype=int)

        if level in ini.names:
            level_idx = ini.names.index(level)
            levels[level_idx] = new_level
            codes[level_idx] = new_codes
        else:
            levels.append(new_level)
            codes.append(new_codes)
            names.append(level)

    res = pd.MultiIndex(levels=levels, codes=codes, names=names)

    return res


def set_index_levels_func(
    pobj: P,
    levels_to_set: dict[str, Any | Collection[Any]],
    copy: bool = True,
) -> P:
    """
    Set the index levels of a [pd.DataFrame][pandas.DataFrame]

    Parameters
    ----------
    pobj
        [pd.DataFrame][pandas.DataFrame] to update

    levels_to_set
        Mapping of level names to values to set

    copy
        Should `pobj` be copied before returning?

    Returns
    -------
    :
        `pobj` with updates applied to its index
    """
    if not isinstance(pobj.index, pd.MultiIndex):
        msg = (
            "This function is only intended to be used "
            "when `df`'s index is an instance of `MultiIndex`. "
            f"Received {type(pobj.index)=}"
        )
        raise TypeError(msg)

    if copy:
        pobj = pobj.copy()

    pobj.index = set_levels(pobj.index, levels_to_set=levels_to_set)  # type: ignore

    return pobj



"""
Test `pandas_openscm.index_manipulation.ensure_index_is_multiindex`
"""

from __future__ import annotations

import numpy as np
import pandas as pd
import pytest

from pandas_openscm.index_manipulation import ensure_index_is_multiindex
from pandas_openscm.testing import create_test_df


@pytest.mark.parametrize("copy, copy_exp", ((None, True), (True, True), (False, False)))
def test_ensure_index_is_multiindex(copy, copy_exp):
    start = pd.DataFrame(
        [[1, 2], [3, 4]],
        columns=[10, 20],
        index=pd.Index(["a", "b"], name="variable"),
    )

    call_kwargs = {}
    if copy is not None:
        call_kwargs["copy"] = copy

    res = ensure_index_is_multiindex(start, **call_kwargs)

    assert isinstance(res.index, pd.MultiIndex)
    assert res.index.names == ["variable"]

    if copy_exp:
        # New object returned
        assert id(start) != id(res)

        assert isinstance(start.index, pd.Index)

    else:
        # Same object returned
        assert id(start) == id(res)
        # Therefore affects input object too
        assert isinstance(start.index, pd.MultiIndex)


@pytest.mark.parametrize("copy, copy_exp", ((None, True), (True, True), (False, False)))
def test_ensure_index_is_multiindex_no_op(copy, copy_exp):
    start = create_test_df(
        variables=[("Temperature", "K")],
        n_scenarios=2,
        n_runs=2,
        timepoints=np.arange(1750.0, 2100.0 + 1.0),
    )

    call_kwargs = {}
    if copy is not None:
        call_kwargs["copy"] = copy

    res = ensure_index_is_multiindex(start, **call_kwargs)

    # Already a MultiIndex, should be no change
    pd.testing.assert_index_equal(res.index, start.index)

    # Behaviour of copy should be respected to avoid confusing behaviour
    if copy_exp:
        # New object returned
        assert id(start) != id(res)

    else:
        # Same object returned
        assert id(start) == id(res)


@pytest.mark.parametrize("copy, copy_exp", ((None, True), (True, True), (False, False)))
def test_accessor(setup_pandas_accessors, copy, copy_exp):
    start = pd.DataFrame(
        [[1, 2], [3, 4]],
        columns=[10, 20],
        index=pd.Index(["a", "b"], name="variable"),
    )

    call_kwargs = {}
    if copy is not None:
        call_kwargs["copy"] = copy

    res = start.openscm.ensure_index_is_multiindex(**call_kwargs)

    assert isinstance(res.index, pd.MultiIndex)
    assert res.index.names == ["variable"]
    if copy_exp:
        # New object returned
        assert id(start) != id(res)

    else:
        # Same object returned
        assert id(start) == id(res)

    # Test alias too
    res_short = start.openscm.eiim(**call_kwargs)

    assert isinstance(res_short.index, pd.MultiIndex)
    assert res_short.index.names == ["variable"]
    if copy_exp:
        # New object returned
        assert id(start) != id(res_short)

    else:
        # Same object returned
        assert id(start) == id(res_short)



"""
Serialisation/deserialisation (i.e. input/output) support
"""

from __future__ import annotations

from pathlib import Path

import pandas as pd


def load_timeseries_csv(
    fp: Path,
    lower_column_names: bool = True,
    index_columns: list[str] | None = None,
    out_columns_type: type | None = None,
    out_columns_name: str | None = None,
) -> pd.DataFrame:
    """
    Load a CSV holding timeseries

    In other words, a CSV that has metadata columns
    and then some time columns.

    Parameters
    ----------
    fp
        File path to load

    lower_column_names
        Convert the column names to all lower case as part of loading.

        Note, if `lower_col_names` is `True`,
        the column names are converted to lower case
        before the index is set so
        a) you should only use lower case in `index_columns`
        and b) the lowering will affect values that do not end up in the index too.

    index_columns
        Columns to treat as metadata from the loaded CSV.

        At the moment, if this is not provided, a `NotImplementedError` is raised.
        In future, if not provided, we will try and infer the columns
        based on whether they look like time columns or not.

    out_columns_type
        The type to apply to the output columns that are not part of the index.

        If not supplied, the raw type returned by pandas is returned.

    out_columns_name
        The name for the columns in the output.

        If not supplied, the raw name returned by pandas is returned.

        This can also be set with
        [pd.DataFrame.rename_axis][pandas.DataFrame.rename_axis]
        but we provide it here for convenience
        (and in case you couldn't find this trick for ages, like us).

    Returns
    -------
    :
        Loaded data
    """
    out = pd.read_csv(fp)

    if lower_column_names:
        out.columns = out.columns.str.lower()

    if index_columns is None:
        raise NotImplementedError(index_columns)

    out = out.set_index(index_columns)

    if out_columns_type is not None:
        out.columns = out.columns.astype(out_columns_type)

    if out_columns_name is not None:
        out = out.rename_axis(out_columns_name, axis="columns")

    return out



"""
API for [pandas][] accessors.

Accessors for [pd.DataFrame][pandas.DataFrame]'s,
[pd.Series][pandas.Series]'s
and [pd.Index][pandas.Index]'s.

**Notes for developers**

We try and keep the accessors as a super-thin layer.
This makes it easier to re-use functionality in a functional way,
which is beneficial
(particularly if we one day need to switch to
a different kind of dataframe e.g. dask).

As a result, we effectively duplicate our API in the accessor layer.
This is ok for now, because this repo is not so big.
Pandas and pandas-indexing use pandas' `pandas.util._decorators.docs` decorator
(see https://github.com/pandas-dev/pandas/blob/05de25381f71657bd425d2c4045d81a46b2d3740/pandas/util/_decorators.py#L342)
to avoid duplicating the docs.
We could use the same pattern, but I have found that this magic
almost always goes wrong so I would stay away from this as long as we can.

We would like to move to a less error-prone, less manual solution.
We tried using mix-ins, but this is just a yuck pattern
that makes it really hard to see where functionality comes from
(a common issue with inheritance)
and makes the type hinting hard.
As a result, we aren't using it.

Probably the next thing to try is auto-generating the code from some template.
This is basically the same idea as using a macro in C.
It likely wouldn't be that hard, and would be much more robust.
"""

from __future__ import annotations

import pandas as pd

from pandas_openscm.accessors.dataframe import PandasDataFrameOpenSCMAccessor
from pandas_openscm.accessors.series import PandasSeriesOpenSCMAccessor


def register_pandas_accessors(namespace: str = "openscm") -> None:
    """
    Register the pandas accessors

    This registers accessors
    for [DataFrame][pandas.DataFrame]'s, [Series][pandas.Series]'s
    and [Index][pandas.Index]'s.
    If you only want to register accessors for one of these,
    we leave it up to you to copy the line(s) you need.

    For details of how these accessors work, see
    [pandas' docs](https://pandas.pydata.org/docs/development/extending.html#registering-custom-accessors).

    We provide this as a separate function
    because we have had really bad experiences with imports having side effects
    (which seems to be the more normal pattern)
    and don't want to pass those bad experiences on.

    Parameters
    ----------
    namespace
        Namespace to use for the accessor

        E.g. if namespace is 'custom'
        then the pandas-openscm API will be available under
        `pd.DataFrame.custom.pandas_openscm_function`
        e.g. `pd.DataFrame.custom.convert_unit`.
    """
    pd.api.extensions.register_dataframe_accessor(namespace)(
        PandasDataFrameOpenSCMAccessor
    )
    pd.api.extensions.register_series_accessor(namespace)(PandasSeriesOpenSCMAccessor)
    # pd.api.extensions.register_index_accessor(namespace)(PandasIndexOpenSCMAccessor)



"""
Tools for reshaping data in common ways
"""

from __future__ import annotations

import pandas as pd


def ts_to_long_data(df: pd.DataFrame, time_col_name: str = "time") -> pd.DataFrame:
    """
    Convert timeseries data to long data

    Parameters
    ----------
    df
        Data to convert

    time_col_name
        Name of the time column in the output

    Returns
    -------
    :
        `df` in long-form
    """
    return df.melt(ignore_index=False, var_name=time_col_name).reset_index()



"""
Accessor for [pd.Series][pandas.Series]
"""

from __future__ import annotations

from collections.abc import Collection, Mapping
from typing import TYPE_CHECKING, Any, Generic, TypeVar

import pandas as pd

from pandas_openscm.index_manipulation import (
    set_index_levels_func,
)
from pandas_openscm.unit_conversion import convert_unit, convert_unit_like

if TYPE_CHECKING:
    # Hmm this is somehow not correct.
    # Figuring it out is a job for another day
    S = TypeVar("S", bound=pd.Series[Any])

    import pint

else:
    S = TypeVar("S")


class PandasSeriesOpenSCMAccessor(Generic[S]):
    """
    [pd.Series][pandas.Series] accessor

    For details, see
    [pandas' docs](https://pandas.pydata.org/docs/development/extending.html#registering-custom-accessors).
    """

    def __init__(self, series: S):
        """
        Initialise

        Parameters
        ----------
        series
            [pd.Series][pandas.Series] to use via the accessor
        """
        # It is possible to validate here.
        # However, it's probably better to do validation closer to the data use.
        self._series = series

    def convert_unit(
        self,
        desired_units: str | Mapping[str, str] | pd.Series[str],
        unit_level: str = "unit",
        ur: pint.facets.PlainRegistry | None = None,
    ) -> S:
        """
        Convert units

        This uses [convert_unit_from_target_series][pandas_openscm.unit_conversion.].
        If you want to understand the details of how the conversion works,
        see that function's docstring.

        Parameters
        ----------
        desired_units
            Desired unit(s) for `series`

            If this is a string,
            we attempt to convert all timeseries to the given unit.

            If this is a mapping,
            we convert the given units to the target units.
            Be careful using this form - you need to be certain of the units.
            If any of your keys don't match the existing units
            (even by a single whitespace character)
            then the unit conversion will not happen.

            If this is a [pd.Series][pandas.Series],
            then it will be passed to
            [convert_unit_from_target_series][pandas_openscm.unit_conversion.]
            after filling any rows in the [pd.Series][pandas.Series]
            that are not in `desired_units`
            with the existing unit (i.e. unspecified rows are not converted).

            For further details, see the examples
            in [convert_unit][pandas_openscm.unit_conversion.].

        unit_level
            Level in the index which holds unit information

            Passed to
            [convert_unit_from_target_series][pandas_openscm.unit_conversion.].

        ur
            Unit registry to use for the conversion.

            Passed to
            [convert_unit_from_target_series][pandas_openscm.unit_conversion.].

        Returns
        -------
        :
            Data with converted units
        """
        res = convert_unit(
            self._series, desired_units=desired_units, unit_level=unit_level, ur=ur
        )

        # The type hinting is impossible to get right here
        # because the casting doesn't work to match the return type
        # (the return type is the same as the input,
        # but we would have to cast to make sure it's numeric
        # and we can't do a runtime check because pd.Series
        # is not subscriptable at runtime).
        # Hence just ignore the type stuff,
        # it's impossible to get right with pandas' accessor pattern.
        # If users want correct type hints, they should use the functional form.
        return res  # type: ignore

    def convert_unit_like(
        self,
        target: pd.DataFrame | pd.Series[Any],
        unit_level: str = "unit",
        target_unit_level: str | None = None,
        ur: pint.facets.PlainRegistry | None = None,
    ) -> S:
        """
        Convert units to match another supported pandas object

        For further details, see the examples
        in [convert_unit_like][pandas_openscm.unit_conversion.].

        This is essentially a helper for
        [convert_unit_from_target_series][pandas_openscm.unit_conversion.].
        It implements one set of logic for extracting desired units
        and tries to be clever, handling differences in index levels
        between the data and `target` sensibly wherever possible.

        If you want behaviour other than what is implemented here,
        use [convert_unit_from_target_series][pandas_openscm.unit_conversion.] directly.

        Parameters
        ----------
        target
            Supported [pandas][] object whose units should be matched

        unit_level
            Level in the data's index which holds unit information

        target_unit_level
            Level in `target`'s index which holds unit information

            If not supplied, we use `unit_level`.

        ur
            Unit registry to use for the conversion.

            Passed to
            [convert_unit_from_target_series][pandas_openscm.unit_conversion.].

        Returns
        -------
        :
            Data with converted units
        """
        res = convert_unit_like(
            self._series,
            target=target,
            unit_level=unit_level,
            target_unit_level=target_unit_level,
            ur=ur,
        )

        # The type hinting is impossible to get right here
        # because the casting doesn't work to match the return type
        # (the return type is the same as the input,
        # but we would have to cast to make sure it's numeric
        # and we can't do a runtime check because pd.Series
        # is not subscriptable at runtime).
        # Hence just ignore the type stuff,
        # it's impossible to get right with pandas' accessor pattern.
        # If users want correct type hints, they should use the functional form.
        return res  # type: ignore

    # def ensure_index_is_multiindex(self, copy: bool = True) -> pd.DataFrame:
    #     """
    #     Ensure that the index is a [pd.MultiIndex][pandas.MultiIndex]
    #
    #     Parameters
    #     ----------
    #     copy
    #         Whether to copy `df` before manipulating the index name
    #
    #     Returns
    #     -------
    #     :
    #         `df` with a [pd.MultiIndex][pandas.MultiIndex]
    #
    #         If the index was already a [pd.MultiIndex][pandas.MultiIndex],
    #         this is a no-op (although the value of copy is respected).
    #     """
    #     return ensure_index_is_multiindex(self._df, copy=copy)
    #
    # def eiim(self, copy: bool = True) -> pd.DataFrame:
    #     """
    #     Ensure that the index is a [pd.MultiIndex][pandas.MultiIndex]
    #
    #     Alias for [ensure_index_is_multiindex][pandas_openscm.index_manipulation.]
    #
    #     Parameters
    #     ----------
    #     copy
    #         Whether to copy `df` before manipulating the index name
    #
    #     Returns
    #     -------
    #     :
    #         `df` with a [pd.MultiIndex][pandas.MultiIndex]
    #
    #         If the index was already a [pd.MultiIndex][pandas.MultiIndex],
    #         this is a no-op (although the value of copy is respected).
    #     """
    #     return self.ensure_index_is_multiindex(copy=copy)
    #
    # def fix_index_name_after_groupby_quantile(
    #     self, new_name: str = "quantile", copy: bool = False
    # ) -> pd.DataFrame:
    #     """
    #     Fix the index name after performing a `groupby(...).quantile(...)` operation
    #
    #     By default, pandas doesn't assign a name to the quantile level
    #     when doing an operation of the form given above.
    #     This fixes this, but it does assume
    #     that the quantile level is the only unnamed level in the index.
    #
    #     Parameters
    #     ----------
    #     new_name
    #         New name to give to the quantile column
    #
    #     copy
    #         Whether to copy `df` before manipulating the index name
    #
    #     Returns
    #     -------
    #     :
    #         `df`, with the last level in its index renamed to `new_name`.
    #     """
    #     return fix_index_name_after_groupby_quantile(
    #         self._df, new_name=new_name, copy=copy
    #     )
    #
    # def groupby_except(
    #     self, non_groupers: str | list[str], observed: bool = True
    # ) -> pd.core.groupby.generic.DataFrameGroupBy[Any]:
    #     """
    #     Group by all index levels except specified levels
    #
    #     This is the inverse of [pd.DataFrame.groupby][pandas.DataFrame.groupby].
    #
    #     Parameters
    #     ----------
    #     non_groupers
    #         Columns to exclude from the grouping
    #
    #     observed
    #         Whether to only return observed combinations or not
    #
    #     Returns
    #     -------
    #     :
    #         The [pd.DataFrame][pandas.DataFrame],
    #         grouped by all columns except `non_groupers`.
    #     """
    #     return groupby_except(df=self._df, non_groupers=non_groupers, observed=observed)  # noqa: E501
    #
    # def mi_loc(
    #     self,
    #     locator: pd.Index[Any] | pd.MultiIndex | pix.selectors.Selector,
    # ) -> pd.DataFrame:
    #     """
    #     Select data, being slightly smarter than the default [pandas.DataFrame.loc][].
    #
    #     Parameters
    #     ----------
    #     locator
    #         Locator to apply
    #
    #         If this is a multi-index, we use
    #         [multi_index_lookup][pandas_openscm.indexing.] to ensure correct alignment.  # noqa: E501
    #
    #         If this is an index that has a name,
    #         we use the name to ensure correct alignment.
    #
    #     Returns
    #     -------
    #     :
    #         Selected data
    #
    #     Notes
    #     -----
    #     If you have [pandas_indexing][] installed,
    #     you can get the same (perhaps even better) functionality
    #     using something like the following instead
    #
    #     ```python
    #     ...
    #     pandas_obj.loc[pandas_indexing.isin(locator)]
    #     ...
    #     ```
    #     """
    #     return mi_loc(self._df, locator)

    def set_index_levels(
        self,
        levels_to_set: dict[str, Any | Collection[Any]],
        copy: bool = True,
    ) -> S:
        """
        Set the index levels

        Parameters
        ----------
        levels_to_set
            Mapping of level names to values to set

        copy
            Should the [pd.Series][pandas.Series] be copied before returning?

        Returns
        -------
        :
            [pd.Series][pandas.Series] with updates applied to its index
        """
        res = set_index_levels_func(
            self._series,
            levels_to_set=levels_to_set,
            copy=copy,
        )

        # Ignore return type
        # because I've done something wrong with how I've set this up.
        # Figuring this out is a job for another day
        return res  # type: ignore

    # def to_category_index(self) -> pd.DataFrame:
    #     """
    #     Convert the index's values to categories
    #
    #     This can save a lot of memory and improve the speed of processing.
    #     However, it comes with some pitfalls.
    #     For a nice discussion of some of them,
    #     see [this article](https://towardsdatascience.com/staying-sane-while-adopting-pandas-categorical-datatypes-78dbd19dcd8a/).
    #
    #     Returns
    #     -------
    #     :
    #         [pd.DataFrame][pandas.DataFrame] with all index columns
    #         converted to category type.
    #     """
    #     return convert_index_to_category_index(self._df)

    # def update_index_levels(
    #     self,
    #     updates: dict[Any, Callable[[Any], Any]],
    #     copy: bool = True,
    #     remove_unused_levels: bool = True,
    # ) -> pd.DataFrame:
    #     """
    #     Update the index levels
    #
    #     Parameters
    #     ----------
    #     updates
    #         Updates to apply to the index levels
    #
    #         Each key is the index level to which the updates will be applied.
    #         Each value is a function which updates the levels to their new values.
    #
    #     copy
    #         Should the [pd.DataFrame][pandas.DataFrame] be copied before returning?
    #
    #     remove_unused_levels
    #         Remove unused levels before applying the update
    #
    #         Specifically, call
    #         [pd.MultiIndex.remove_unused_levels][pandas.MultiIndex.remove_unused_levels].  # noqa: E501
    #
    #         This avoids trying to update levels that aren't being used.
    #
    #     Returns
    #     -------
    #     :
    #         [pd.DataFrame][pandas.DataFrame] with updates applied to its index
    #     """
    #     return update_index_levels_func(
    #         self._df,
    #         updates=updates,
    #         copy=copy,
    #         remove_unused_levels=remove_unused_levels,
    #     )
    #
    # def update_index_levels_from_other(
    #     self,
    #     update_sources: dict[
    #         Any, tuple[Any, Callable[[Any], Any] | dict[Any, Any] | pd.Series[Any]]
    #     ],
    #     copy: bool = True,
    #     remove_unused_levels: bool = True,
    # ) -> pd.DataFrame:
    #     """
    #     Update the index levels based on other index levels
    #
    #     Parameters
    #     ----------
    #     update_sources
    #         Updates to apply to `df`'s index
    #
    #         Each key is the level to which the updates will be applied
    #         (or the level that will be created if it doesn't already exist).
    #
    #         Each value is a tuple of which the first element
    #         is the level to use to generate the values (the 'source level')
    #         and the second is mapper of the form used by
    #         [pd.Index.map][pandas.Index.map]
    #         which will be applied to the source level
    #         to update/create the level of interest.
    #
    #     copy
    #         Should the [pd.DataFrame][pandas.DataFrame] be copied before returning?
    #
    #     remove_unused_levels
    #         Remove unused levels before applying the update
    #
    #         Specifically, call
    #         [pd.MultiIndex.remove_unused_levels][pandas.MultiIndex.remove_unused_levels].  # noqa: E501
    #
    #         This avoids trying to update levels that aren't being used.
    #
    #     Returns
    #     -------
    #     :
    #         [pd.DataFrame][pandas.DataFrame] with updates applied to its index
    #     """
    #     return update_index_levels_from_other_func(
    #         self._df,
    #         update_sources=update_sources,
    #         copy=copy,
    #         remove_unused_levels=remove_unused_levels,
    #     )



"""
Tests of `pandas_openscm.db.path_handling`
"""

import re
from contextlib import nullcontext as does_not_raise
from pathlib import Path

import pytest

from pandas_openscm.db.path_handling import DBPath


@pytest.mark.parametrize(
    "abs, rel_db, exp",
    (
        (Path("/a/b/c/d.csv"), Path("d.csv"), does_not_raise()),
        (Path("/a/b/c/d.csv"), Path("c/d.csv"), does_not_raise()),
        (
            Path("/a/b/c/d.csv"),
            Path("e/d.csv"),
            pytest.raises(
                AssertionError,
                match="".join(
                    (
                        re.escape("rel_db value, "),
                        ".*Path",
                        re.escape(r"('e/d.csv'), is not a sub-path of self.abs="),
                        ".*Path",
                        re.escape("('/a/b/c/d.csv')"),
                    )
                ),
            ),
        ),
    ),
)
def test_rel_db_validator(abs, rel_db, exp):
    with exp:
        DBPath(abs=abs, rel_db=rel_db)



"""
Test `pandas_openscm.index_manipulation.update_levels`
"""

from __future__ import annotations

import re

import numpy as np
import pandas as pd
import pytest

from pandas_openscm.index_manipulation import update_index_levels_func, update_levels


@pytest.mark.parametrize(
    "start, updates",
    (
        pytest.param(
            pd.MultiIndex.from_tuples(
                [
                    ("sa", "va", "kg", 0),
                    ("sb", "vb", "m", 1),
                    ("sa", "va", "kg", 2),
                ],
                names=["scenario", "variable", "unit", "run_id"],
            ),
            {},
            id="no-changes",
        ),
        pytest.param(
            pd.MultiIndex.from_tuples(
                [
                    ("sa", "va", "kg", 0),
                    ("sb", "vb", "m", 1),
                    ("sa", "va", "kg", 2),
                ],
                names=["scenario", "variable", "unit", "run_id"],
            ),
            {"variable": lambda x: x.replace("v", "vv")},
            id="single-update",
        ),
        pytest.param(
            pd.MultiIndex.from_tuples(
                [
                    ("sa", "va", "kg", 0),
                    ("sb", "vb", "m", 1),
                    ("sa", "va", "kg", 2),
                ],
                names=["scenario", "variable", "unit", "run_id"],
            ),
            {
                "variable": lambda x: x.replace("v", "vv"),
                "unit": lambda x: x.replace("kg", "g").replace("m", "km"),
            },
            id="multiple-updates",
        ),
        pytest.param(
            pd.MultiIndex.from_tuples(
                [
                    ("sa", "va", "kg", 0),
                    ("sb", "vb", "m", -1),
                    ("sa", "va", "kg", -2),
                    ("sa", "vb", "kg", 2),
                ],
                names=["scenario", "variable", "unit", "run_id"],
            ),
            {"variable": lambda x: x[0]},
            id="updates-lead-to-dups",
        ),
        pytest.param(
            pd.MultiIndex.from_tuples(
                [
                    ("sa", "va", "kg", 0),
                    ("sb", "vb", "m", -1),
                    ("sa", "va", "kg", -2),
                    ("sa", "vb", "kg", 2),
                ],
                names=["scenario", "variable", "unit", "run_id"],
            ),
            {
                "variable": lambda x: x.replace("v", "vv"),
                "unit": lambda x: x.replace("kg", "g").replace("m", "km"),
                "run_id": np.abs,
            },
            id="multiple-updates-incl-external-func",
        ),
    ),
)
def test_update_levels(start, updates):
    res = update_levels(start, updates=updates)

    exp = start.to_frame(index=False)
    for level, func in updates.items():
        exp[level] = exp[level].map(func)
    exp = pd.MultiIndex.from_frame(exp)

    pd.testing.assert_index_equal(res, exp)


def test_update_levels_missing_level():
    start = pd.MultiIndex.from_tuples(
        [
            ("sa", "va", "kg", 0),
            ("sb", "vb", "m", -1),
            ("sa", "va", "kg", -2),
            ("sa", "vb", "kg", 2),
        ],
        names=["scenario", "variable", "unit", "run_id"],
    )
    updates = {
        "variable": lambda x: x.replace("v", "vv"),
        "units": lambda x: x.replace("kg", "g").replace("m", "km"),
    }

    with pytest.raises(
        KeyError,
        match=re.escape(
            "units is not available in the index. "
            f"Available levels: {['scenario', 'variable', 'unit', 'run_id']}"
        ),
    ):
        update_levels(start, updates=updates)


def test_doesnt_trip_over_droped_levels(setup_pandas_accessors):
    def update_func(in_v: int) -> int:
        if in_v < 0:
            msg = f"Value must be greater than zero, received {in_v}"
            raise ValueError(msg)

        return in_v * -1

    start = pd.MultiIndex.from_tuples(
        [
            ("sa", "va", "kg", 0),
            ("sb", "vb", "m", 1),
            ("sa", "va", "kg", 2),
            ("sa", "vb", "kg", -2),
        ],
        names=["scenario", "variable", "unit", "run_id"],
    )

    updates = {"run_id": update_func}

    res = update_levels(start[:-1], updates=updates)

    exp = pd.MultiIndex.from_tuples(
        [
            ("sa", "va", "kg", 0),
            ("sb", "vb", "m", -1),
            ("sa", "va", "kg", -2),
        ],
        names=["scenario", "variable", "unit", "run_id"],
    )
    pd.testing.assert_index_equal(res, exp)

    # If you turn the drop off, you get an error
    exp_error_no_removal = pytest.raises(
        ValueError, match=re.escape("Value must be greater than zero, received -2")
    )
    with exp_error_no_removal:
        # Even though we're not using the levels,
        # they still get mapped if we don't remove them
        update_levels(start[:-1], updates=updates, remove_unused_levels=False)

    # Same thing but from a DataFrame
    start_df = pd.DataFrame(
        np.zeros((start.shape[0], 3)), columns=[2010, 2020, 2030], index=start
    )

    res_df = update_index_levels_func(start_df.iloc[:-1, :], updates=updates)

    exp_df = pd.DataFrame(
        np.zeros((exp.shape[0], 3)), columns=start_df.columns, index=exp
    )

    pd.testing.assert_frame_equal(res_df, exp_df)
    with exp_error_no_removal:
        update_index_levels_func(
            start_df.iloc[:-1, :], updates=updates, remove_unused_levels=False
        )

    # Lastly, test the accessor
    pd.testing.assert_frame_equal(
        start_df.iloc[:-1, :].openscm.update_index_levels(updates), exp_df
    )
    with exp_error_no_removal:
        start_df.iloc[:-1, :].openscm.update_index_levels(
            updates, remove_unused_levels=False
        )


def test_accessor(setup_pandas_accessors):
    start = pd.DataFrame(
        np.arange(2 * 4).reshape((4, 2)),
        columns=[2010, 2020],
        index=pd.MultiIndex.from_tuples(
            [
                ("sa", "va", "kg", 0),
                ("sb", "vb", "m", -1),
                ("sa", "va", "kg", -2),
                ("sa", "vb", "kg", 2),
            ],
            names=["scenario", "variable", "unit", "run_id"],
        ),
    )

    updates = {
        "variable": lambda x: x.replace("v", "vv"),
        "unit": lambda x: x.replace("kg", "g").replace("m", "km"),
    }

    exp = pd.DataFrame(
        start.values,
        columns=start.columns,
        index=pd.MultiIndex.from_tuples(
            [
                ("sa", "vva", "g", 0),
                ("sb", "vvb", "km", -1),
                ("sa", "vva", "g", -2),
                ("sa", "vvb", "g", 2),
            ],
            names=["scenario", "variable", "unit", "run_id"],
        ),
    )

    res = start.openscm.update_index_levels(updates)
    pd.testing.assert_frame_equal(res, exp)

    # Test function too
    res = update_index_levels_func(start, updates)
    pd.testing.assert_frame_equal(res, exp)


def test_accessor_not_multiindex(setup_pandas_accessors):
    start = pd.DataFrame(np.arange(2 * 4).reshape((4, 2)))

    error_msg = re.escape(
        "This function is only intended to be used "
        "when `df`'s index is an instance of `MultiIndex`. "
        "Received type(df.index)="
    )
    with pytest.raises(TypeError, match=error_msg):
        start.openscm.update_index_levels({})

    with pytest.raises(TypeError, match=error_msg):
        update_index_levels_func(start, {})



"""
Integration tests of `pandas_openscm.unit_conversion`
"""

from __future__ import annotations

import re
import sys
from typing import TYPE_CHECKING, Any, TypeVar
from unittest.mock import patch

import numpy as np
import pandas as pd
import pytest

from pandas_openscm.exceptions import MissingOptionalDependencyError
from pandas_openscm.index_manipulation import (
    set_index_levels_func,
)
from pandas_openscm.testing import (
    assert_frame_alike,
    check_result,
    convert_to_desired_type,
    create_test_df,
)
from pandas_openscm.unit_conversion import (
    AmbiguousTargetUnitError,
    MissingDesiredUnitError,
    convert_unit,
    convert_unit_from_target_series,
    convert_unit_like,
)

if TYPE_CHECKING:
    P = TypeVar("P", pd.DataFrame | pd.Series[Any])

pobj_type = pytest.mark.parametrize(
    "pobj_type",
    ("DataFrame", "Series"),
)
"""
Parameterisation to use to check handling of both DataFrame and Series
"""

check_auto_index_casting_pobj = pytest.mark.parametrize(
    "only_two_index_levels_pobj",
    (
        pytest.param(True, id="only_two_index_levels"),
        pytest.param(False, id="more_than_two_index_levels"),
    ),
)
"""
Parameterisation to use to check handling of auto casting to `pd.Index`

This casting causes all sorts of indexing and other issues.
This parameterisation ensures that we check this edge case.
"""


@pobj_type
@check_auto_index_casting_pobj
def test_convert_unit_no_op(only_two_index_levels_pobj, pobj_type):
    start = create_test_df(
        variables=[
            ("Cold", "mK"),
            ("Warm", "kK"),
            ("Body temperature", "degC"),
        ],
        n_scenarios=2,
        n_runs=3,
        timepoints=np.array([1.0, 2.0, 3.0]),
    )
    if only_two_index_levels_pobj:
        start = start.loc[
            (start.index.get_level_values("scenario") == "scenario_0")
            & (start.index.get_level_values("run") == 0)
        ].reset_index(["scenario", "run"], drop=True)

    start = convert_to_desired_type(start, pobj_type)

    res = convert_unit(
        start, start.index.to_frame()["unit"].reset_index("unit", drop=True)
    )

    check_result(res, start)


def test_convert_unit_unknown_mapping_type():
    start = create_test_df(
        variables=[
            ("Cold", "mK"),
            ("Warm", "kK"),
            ("Body temperature", "degC"),
        ],
        n_scenarios=2,
        n_runs=3,
        timepoints=np.array([1.0, 2.0, 3.0]),
    )

    with pytest.raises(NotImplementedError, match="DataFrame"):
        convert_unit(
            start,
            # DataFrame not supported, has to be Series
            start.index.to_frame()[["unit"]].reset_index("unit", drop=True),
        )


@pobj_type
@check_auto_index_casting_pobj
@pytest.mark.parametrize(
    "unit_level, exp_unit_level",
    (
        pytest.param(None, "unit", id="default"),
        ("units", "units"),
    ),
)
def test_convert_unit_single_unit(
    unit_level, exp_unit_level, only_two_index_levels_pobj, pobj_type
):
    pytest.importorskip("pint")

    start = create_test_df(
        variables=[
            ("Cold", "mK"),
            ("Warm", "kK"),
            ("Body temperature", "degC"),
        ],
        n_scenarios=2,
        n_runs=3,
        timepoints=np.array([1.0, 2.0, 3.0]),
    )
    if only_two_index_levels_pobj:
        start = start.loc[
            (start.index.get_level_values("scenario") == "scenario_0")
            & (start.index.get_level_values("run") == 0)
        ].reset_index(["scenario", "run"], drop=True)

    start = convert_to_desired_type(start, pobj_type)

    call_kwargs = {}
    if unit_level is not None:
        start = start.rename_axis(index={"unit": unit_level})
        call_kwargs["unit_level"] = unit_level

    res = convert_unit(start, "K", **call_kwargs)

    assert (res.index.get_level_values(exp_unit_level) == "K").all()

    np.testing.assert_equal(
        res.loc[res.index.get_level_values("variable") == "Cold", :].values,
        1e-3 * start.loc[start.index.get_level_values("variable") == "Cold", :].values,
    )

    np.testing.assert_equal(
        res.loc[res.index.get_level_values("variable") == "Warm", :].values,
        1e3 * start.loc[start.index.get_level_values("variable") == "Warm", :].values,
    )

    np.testing.assert_equal(
        res.loc[res.index.get_level_values("variable") == "Body temperature", :].values,
        273.15
        + start.loc[
            start.index.get_level_values("variable") == "Body temperature", :
        ].values,
    )


def test_convert_unit_ur_injection():
    pint = pytest.importorskip("pint")

    start = create_test_df(
        variables=[("Wavelength", "m")],
        n_scenarios=2,
        n_runs=2,
        timepoints=np.array([1.0, 2.0, 3.0]),
    )

    # Without injection, raises
    with pytest.raises(pint.DimensionalityError):
        convert_unit(start, "Hz")

    # With injection and context, all good
    ur = pint.UnitRegistry()
    with ur.context("spectroscopy"):
        res = convert_unit(start, "Hz", ur=ur)

    np.testing.assert_allclose(
        res.values,
        2.998 * 1e8 / start.values,
        rtol=1e-4,
    )


@pobj_type
@check_auto_index_casting_pobj
def test_convert_unit_mapping(only_two_index_levels_pobj, pobj_type):
    pytest.importorskip("pint")

    start = create_test_df(
        variables=[
            ("temperature", "K"),
            ("erf", "W / m^2"),
            ("ohc", "ZJ"),
        ],
        n_scenarios=2,
        n_runs=2,
        timepoints=np.array([1850.0, 2000.0, 2050.0, 2100.0]),
    )
    if only_two_index_levels_pobj:
        start = start.loc[
            (start.index.get_level_values("scenario") == "scenario_0")
            & (start.index.get_level_values("run") == 0)
        ].reset_index(["scenario", "run"], drop=True)

    start = convert_to_desired_type(start, pobj_type)

    # Don't convert W / m^2
    res = convert_unit(start, {"K": "degC", "ZJ": "J"})

    np.testing.assert_allclose(
        res.loc[res.index.get_level_values("variable") == "temperature", :].values,
        -273.15
        + start.loc[
            start.index.get_level_values("variable") == "temperature", :
        ].values,
    )

    np.testing.assert_equal(
        res.loc[res.index.get_level_values("variable") == "erf", :].values,
        start.loc[start.index.get_level_values("variable") == "erf", :].values,
    )

    np.testing.assert_equal(
        res.loc[res.index.get_level_values("variable") == "ohc", :].values,
        1e21 * start.loc[start.index.get_level_values("variable") == "ohc", :].values,
    )


@pobj_type
@check_auto_index_casting_pobj
def test_convert_series(only_two_index_levels_pobj, pobj_type):
    pytest.importorskip("pint")

    # Check that conversion works if user supplies a Series of target units
    start = create_test_df(
        variables=[
            ("temperature", "K"),
            ("erf", "W / m^2"),
            ("ohc", "ZJ"),
        ],
        n_scenarios=2,
        n_runs=2,
        timepoints=np.array([1850.0, 2000.0, 2050.0, 2100.0]),
    )
    if only_two_index_levels_pobj:
        start = start.loc[
            (start.index.get_level_values("scenario") == "scenario_0")
            & (start.index.get_level_values("run") == 0)
        ].reset_index(["scenario", "run"], drop=True)

    start = convert_to_desired_type(start, pobj_type)

    desired_units = (
        start.loc[start.index.get_level_values("variable") != "temperature"]
        .reset_index("unit")["unit"]
        .replace({"W / m^2": "ZJ / yr / m^2", "ZJ": "PJ"})
    )

    res = convert_unit(start, desired_units)

    np.testing.assert_allclose(
        res.loc[res.index.get_level_values("variable") == "temperature", :].values,
        start.loc[start.index.get_level_values("variable") == "temperature", :].values,
    )

    np.testing.assert_allclose(
        res.loc[res.index.get_level_values("variable") == "erf", :].values,
        (60.0 * 60.0 * 24.0 * 365.25)
        * 1e-21
        * start.loc[start.index.get_level_values("variable") == "erf", :].values,
    )

    np.testing.assert_allclose(
        res.loc[res.index.get_level_values("variable") == "ohc", :].values,
        1e6 * start.loc[start.index.get_level_values("variable") == "ohc", :].values,
    )


@pobj_type
@check_auto_index_casting_pobj
def test_convert_series_all_rows(only_two_index_levels_pobj, pobj_type):
    pytest.importorskip("pint")

    start = create_test_df(
        variables=[
            ("temperature", "K"),
            ("erf", "W / m^2"),
            ("ohc", "ZJ"),
        ],
        n_scenarios=2,
        n_runs=2,
        timepoints=np.array([1850.0, 2000.0, 2050.0, 2100.0]),
    )
    if only_two_index_levels_pobj:
        start = start.loc[
            (start.index.get_level_values("scenario") == "scenario_0")
            & (start.index.get_level_values("run") == 0)
        ].reset_index(["scenario", "run"], drop=True)

    start = convert_to_desired_type(start, pobj_type)

    desired_units = start.reset_index("unit")["unit"].replace(
        {"W / m^2": "ZJ / yr / m^2", "ZJ": "PJ"}
    )

    res = convert_unit(start, desired_units)

    np.testing.assert_allclose(
        res.loc[res.index.get_level_values("variable") == "temperature", :].values,
        start.loc[start.index.get_level_values("variable") == "temperature", :].values,
    )

    np.testing.assert_allclose(
        res.loc[res.index.get_level_values("variable") == "erf", :].values,
        (60.0 * 60.0 * 24.0 * 365.25)
        * 1e-21
        * start.loc[start.index.get_level_values("variable") == "erf", :].values,
    )

    np.testing.assert_allclose(
        res.loc[res.index.get_level_values("variable") == "ohc", :].values,
        1e6 * start.loc[start.index.get_level_values("variable") == "ohc", :].values,
    )


@pobj_type
@check_auto_index_casting_pobj
def test_convert_series_extra_rows(only_two_index_levels_pobj, pobj_type):
    pytest.importorskip("pint")

    start = create_test_df(
        variables=[
            ("temperature", "K"),
            ("erf", "W / m^2"),
            ("ohc", "ZJ"),
        ],
        n_scenarios=2,
        n_runs=2,
        timepoints=np.array([1850.0, 2000.0, 2050.0, 2100.0]),
    )
    if only_two_index_levels_pobj:
        start = start.loc[
            (start.index.get_level_values("scenario") == "scenario_0")
            & (start.index.get_level_values("run") == 0)
        ].reset_index(["scenario", "run"], drop=True)

    start = convert_to_desired_type(start, pobj_type)

    desired_units = start.reset_index("unit")["unit"].replace(
        {"W / m^2": "ZJ / yr / m^2", "ZJ": "PJ"}
    )
    # Extra rows that aren't in start, should be ignored and not cause failures
    if only_two_index_levels_pobj:
        desired_units.loc[("carbon")] = "GtC"

    else:
        desired_units.loc[("scenario_2", "temperature", 0)] = "kK"

    res = convert_unit(start, desired_units)

    np.testing.assert_allclose(
        res.loc[res.index.get_level_values("variable") == "temperature", :].values,
        start.loc[start.index.get_level_values("variable") == "temperature", :].values,
    )

    np.testing.assert_allclose(
        res.loc[res.index.get_level_values("variable") == "erf", :].values,
        (60.0 * 60.0 * 24.0 * 365.25)
        * 1e-21
        * start.loc[start.index.get_level_values("variable") == "erf", :].values,
    )

    np.testing.assert_allclose(
        res.loc[res.index.get_level_values("variable") == "ohc", :].values,
        1e6 * start.loc[start.index.get_level_values("variable") == "ohc", :].values,
    )


@pobj_type
@check_auto_index_casting_pobj
def test_convert_unit_like_no_op(only_two_index_levels_pobj, pobj_type):
    start = create_test_df(
        variables=[
            ("Cold", "mK"),
            ("Warm", "kK"),
            ("Body temperature", "degC"),
        ],
        n_scenarios=2,
        n_runs=3,
        timepoints=np.array([1.0, 2.0, 3.0]),
    )
    if only_two_index_levels_pobj:
        start = start.loc[
            (start.index.get_level_values("scenario") == "scenario_0")
            & (start.index.get_level_values("run") == 0)
        ].reset_index(["scenario", "run"], drop=True)

    start = convert_to_desired_type(start, pobj_type)

    res = convert_unit_like(start, start)

    check_result(res, start)


check_auto_index_casting_target = pytest.mark.parametrize(
    "only_two_index_levels_target",
    (
        pytest.param(True, id="only_two_index_levels"),
        pytest.param(False, id="more_than_two_index_levels"),
    ),
)
"""
Parameterisation to use to check handling of auto casting to `pd.Index`

This casting causes all sorts of indexing and other issues.
This parameterisation ensures that we check this edge case.
"""

target_type = pytest.mark.parametrize(
    "target_type",
    (
        pytest.param("DataFrame", id="target_DataFrame"),
        pytest.param("Series", id="target_Series"),
    ),
)
"""
Parameterisation to use to check handling of both DataFrame and Series as the target
"""


@pobj_type
@target_type
@check_auto_index_casting_pobj
@check_auto_index_casting_target
def test_convert_unit_like(
    only_two_index_levels_pobj,
    only_two_index_levels_target,
    target_type,
    pobj_type,
):
    pytest.importorskip("pint")

    create_kwargs = dict(
        n_scenarios=2,
        n_runs=3,
        timepoints=np.array([1.0, 2.0, 3.0]),
    )
    start = create_test_df(
        variables=[
            ("Cold", "mK"),
            ("Warm", "kK"),
            ("Body temperature", "degC"),
        ],
        **create_kwargs,
    )
    if only_two_index_levels_pobj:
        start = start.loc[
            (start.index.get_level_values("scenario") == "scenario_0")
            & (start.index.get_level_values("run") == 0)
        ].reset_index(["scenario", "run"], drop=True)

    start = convert_to_desired_type(start, pobj_type)

    target = create_test_df(
        variables=[
            ("Cold", "microK"),
            ("Warm", "MK"),
            ("Body temperature", "degF"),
        ],
        **create_kwargs,
    )
    if only_two_index_levels_target:
        target = target.loc[
            (target.index.get_level_values("scenario") == "scenario_0")
            & (target.index.get_level_values("run") == 0)
        ].reset_index(["scenario", "run"], drop=True)

    target = convert_to_desired_type(target, target_type)

    res = convert_unit_like(start, target)

    exp = convert_unit(start, {"mK": "microK", "kK": "MK", "degC": "degF"})

    check_result(res, exp)


@pobj_type
@target_type
@check_auto_index_casting_pobj
@check_auto_index_casting_target
def test_convert_unit_like_missing_levels(
    only_two_index_levels_pobj,
    only_two_index_levels_target,
    target_type,
    pobj_type,
):
    pytest.importorskip("pint")

    start = create_test_df(
        variables=[
            ("Cold", "mK"),
            ("Warm", "kK"),
            ("Body temperature", "degC"),
        ],
        n_scenarios=1,
        n_runs=2,
        timepoints=np.array([2020.0, 2030.0, 2040.0]),
    )
    if only_two_index_levels_pobj:
        start = start.loc[
            (start.index.get_level_values("scenario") == "scenario_0")
            & (start.index.get_level_values("run") == 0)
        ].reset_index(["scenario", "run"], drop=True)

    start = convert_to_desired_type(start, pobj_type)

    target = create_test_df(
        variables=[
            ("Cold", "K"),
            ("Warm", "K"),
            ("Body temperature", "degF"),
        ],
        n_scenarios=1,
        n_runs=1,
        timepoints=np.array([1.0, 2.0, 3.0]),
    ).reset_index("run", drop=True)
    if only_two_index_levels_target:
        target = target.reset_index("scenario", drop=True)

    target = convert_to_desired_type(target, target_type)

    res = convert_unit_like(start, target)

    exp = convert_unit(start, {"mK": "K", "kK": "K", "degC": "degF"})

    check_result(res, exp)


@pobj_type
@target_type
@check_auto_index_casting_pobj
@check_auto_index_casting_target
def test_convert_unit_like_missing_specs(
    only_two_index_levels_pobj,
    only_two_index_levels_target,
    target_type,
    pobj_type,
):
    """
    Test conversion when the target doesn't specify a unit for all rows in start
    """
    pytest.importorskip("pint")

    start = create_test_df(
        variables=[
            ("Cold", "mK"),
            ("Warm", "kK"),
            ("Body temperature", "degC"),
        ],
        n_scenarios=1,
        n_runs=2,
        timepoints=np.array([2020.0, 2030.0, 2040.0]),
    )
    if only_two_index_levels_pobj:
        start = start.loc[
            (start.index.get_level_values("scenario") == "scenario_0")
            & (start.index.get_level_values("run") == 0)
        ].reset_index(["scenario", "run"], drop=True)

    start = convert_to_desired_type(start, pobj_type)

    target = create_test_df(
        variables=[
            ("Cold", "K"),
            # ("Warm", "K"),
            ("Body temperature", "degF"),
        ],
        n_scenarios=1,
        n_runs=1,
        timepoints=np.array([1.0, 2.0, 3.0]),
    ).reset_index("run", drop=True)
    if only_two_index_levels_target:
        target = target.reset_index("scenario", drop=True)

    target = convert_to_desired_type(target, target_type)

    res = convert_unit_like(start, target)

    exp = convert_unit(start, {"mK": "K", "degC": "degF"})

    check_result(res, exp)


@pobj_type
@target_type
@check_auto_index_casting_pobj
@check_auto_index_casting_target
def test_convert_unit_like_extra_levels_ok(
    only_two_index_levels_pobj,
    only_two_index_levels_target,
    target_type,
    pobj_type,
):
    pytest.importorskip("pint")

    start = create_test_df(
        variables=[
            ("Cold", "mK"),
            ("Warm", "kK"),
            ("Body temperature", "degC"),
        ],
        n_scenarios=1,
        n_runs=2,
        timepoints=np.array([2020.0, 2030.0, 2040.0]),
    )
    if only_two_index_levels_pobj:
        start = start.loc[
            (start.index.get_level_values("scenario") == "scenario_0")
            & (start.index.get_level_values("run") == 0)
        ].reset_index(["scenario", "run"], drop=True)

    start = convert_to_desired_type(start, pobj_type)

    target = set_index_levels_func(
        create_test_df(
            variables=[
                ("Cold", "K"),
                ("Warm", "K"),
                ("Body temperature", "degF"),
            ],
            n_scenarios=1,
            n_runs=1,
            timepoints=np.array([1.0, 2.0, 3.0]),
        ).reset_index("run", drop=True),
        {"model": "ma"},
    )
    if only_two_index_levels_target:
        target = target.reset_index("scenario", drop=True)

    target = convert_to_desired_type(target, target_type)

    res = convert_unit_like(start, target)

    exp = convert_unit(start, {"mK": "K", "kK": "K", "degC": "degF"})

    check_result(res, exp)


@pobj_type
@target_type
@check_auto_index_casting_pobj
@check_auto_index_casting_target
def test_convert_unit_like_extra_levels_ambiguous_error(
    only_two_index_levels_pobj,
    only_two_index_levels_target,
    target_type,
    pobj_type,
):
    start = create_test_df(
        variables=[
            ("Cold", "mK"),
            ("Warm", "kK"),
            ("Body temperature", "degC"),
        ],
        n_scenarios=1,
        n_runs=2,
        timepoints=np.array([2020.0, 2030.0, 2040.0]),
    )
    if only_two_index_levels_pobj:
        start = start.loc[
            (start.index.get_level_values("scenario") == "scenario_0")
            & (start.index.get_level_values("run") == 0)
        ].reset_index(["scenario", "run"], drop=True)

    start = convert_to_desired_type(start, pobj_type)

    target = pd.DataFrame(
        np.arange(3 * 6).reshape((6, 3)),
        columns=np.array([1.0, 10.0, 100.0]),
        index=pd.MultiIndex.from_tuples(
            [
                ("ma", "scenario_0", "Cold", "microK"),
                ("mb", "scenario_0", "Cold", "K"),
                ("ma", "scenario_0", "Warm", "K"),
                ("mb", "scenario_0", "Warm", "K"),
                ("ma", "scenario_0", "Body temperature", "K"),
                ("mb", "scenario_0", "Body temperature", "degF"),
            ],
            names=["model", "scenario", "variable", "unit"],
        ),
    )
    if only_two_index_levels_target:
        target = target.loc[
            ~(
                (target.index.get_level_values("model") == "ma")
                & (target.index.get_level_values("variable") == "Warm")
            ),
            :,
        ].reset_index(["model", "scenario"], drop=True)

    target = convert_to_desired_type(target, target_type)

    with pytest.raises(AmbiguousTargetUnitError):
        convert_unit_like(start, target)


@pobj_type
@target_type
@check_auto_index_casting_pobj
@check_auto_index_casting_target
def test_convert_unit_like_extra_specs(
    only_two_index_levels_pobj,
    only_two_index_levels_target,
    target_type,
    pobj_type,
):
    """
    Test conversion when the target has a unit for rows that aren't in start
    """
    pytest.importorskip("pint")

    start = create_test_df(
        variables=[
            ("Cold", "mK"),
            ("Warm", "kK"),
            ("Body temperature", "degC"),
        ],
        n_scenarios=1,
        n_runs=2,
        timepoints=np.array([2020.0, 2030.0, 2040.0]),
    )
    if only_two_index_levels_pobj:
        start = start.loc[
            (start.index.get_level_values("scenario") == "scenario_0")
            & (start.index.get_level_values("run") == 0)
        ].reset_index(["scenario", "run"], drop=True)

    start = convert_to_desired_type(start, pobj_type)

    target = create_test_df(
        variables=[
            ("Cold", "K"),
            ("Warm", "K"),
            ("Body temperature", "degF"),
            ("Hot", "PK"),
        ],
        n_scenarios=1,
        n_runs=1,
        timepoints=np.array([1.0, 2.0, 3.0]),
    ).reset_index("run", drop=True)
    if only_two_index_levels_target:
        target = target.reset_index("scenario", drop=True)

    target = convert_to_desired_type(target, target_type)

    res = convert_unit_like(start, target)

    exp = convert_unit(start, {"mK": "K", "kK": "K", "degC": "degF"})

    check_result(res, exp)


def test_convert_unit_like_ur_injection():
    # Importing openscm_units makes the test noticeably slow, hmm
    openscm_units = pytest.importorskip("openscm_units")
    pint = pytest.importorskip("pint")

    start = create_test_df(
        variables=[
            ("emissions_co2", "Mt CO2/yr"),
            ("emissions_n2o", "kt N2O/yr"),
        ],
        n_scenarios=3,
        n_runs=6,
        timepoints=np.array([1.0, 2.0, 3.0]),
    )

    target = create_test_df(
        variables=[
            ("emissions_co2", "Gt C / yr"),
            ("emissions_n2o", "Mt N2ON / yr"),
        ],
        n_scenarios=1,
        n_runs=1,
        timepoints=np.array([1.0, 2.0, 3.0]),
    ).reset_index(["scenario", "run"], drop=True)

    # Without injection, raises
    with pytest.raises(pint.UndefinedUnitError):
        convert_unit_like(start, target)

    # With injection and context, all good
    with openscm_units.unit_registry.context("N2O_conversions"):
        res = convert_unit_like(start, target, ur=openscm_units.unit_registry)

        exp = convert_unit(
            start,
            {"Mt CO2/yr": "Gt C / yr", "kt N2O/yr": "Mt N2ON / yr"},
            ur=openscm_units.unit_registry,
        )

    assert_frame_alike(res, exp)


@pobj_type
@target_type
@pytest.mark.parametrize(
    "unit_level, unit_level_exp, target_unit_level, target_unit_level_exp",
    (
        pytest.param(None, "unit", None, "unit", id="default"),
        pytest.param("units", "units", None, "units", id="target-inferred-from-other"),
        pytest.param("units", "units", "unit", "unit", id="target-other-differ"),
        pytest.param(None, "unit", "units", "units", id="target-specified-only"),
    ),
)
def test_convert_unit_like_unit_level_handling(  # noqa: PLR0913
    unit_level,
    unit_level_exp,
    target_unit_level,
    target_unit_level_exp,
    target_type,
    pobj_type,
):
    pytest.importorskip("pint")

    start = create_test_df(
        variables=[(f"variable_{i}", "kg") for i in range(2)],
        n_scenarios=2,
        n_runs=2,
        timepoints=np.array([1.0, 2.0, 3.0]),
    )

    start = convert_to_desired_type(start, pobj_type)

    target = create_test_df(
        variables=[(f"variable_{i}", "g") for i in range(2)],
        n_scenarios=2,
        n_runs=2,
        timepoints=np.array([10.0, 11.0, 12.0]),
    )

    target = convert_to_desired_type(target, target_type)

    call_kwargs = {}
    if unit_level is not None:
        start = start.rename_axis(index={"unit": unit_level})
        call_kwargs["unit_level"] = unit_level

        if target_unit_level is not None:
            target = target.rename_axis(index={"unit": target_unit_level})
            call_kwargs["target_unit_level"] = target_unit_level

        else:
            target = target.rename_axis(index={"unit": unit_level})

    elif target_unit_level is not None:
        target = target.rename_axis(index={"unit": target_unit_level})
        call_kwargs["target_unit_level"] = target_unit_level

    res = convert_unit_like(start, target, **call_kwargs)

    exp = convert_unit(start, "g", unit_level=unit_level_exp)

    check_result(res, exp)


@pobj_type
def test_convert_unit_from_target_series_missing_desired_unit_error(pobj_type):
    start = pd.DataFrame(
        np.arange(2 * 3).reshape((2, 3)),
        columns=np.array([1.0, 10.0, 100.0]),
        index=pd.MultiIndex.from_tuples(
            [
                ("ma", "Cold", "microK"),
                ("mb", "Cold", "K"),
            ],
            names=["model", "variable", "unit"],
        ),
    )

    start = convert_to_desired_type(start, pobj_type)

    desired_unit = pd.Series(
        [
            "K",
            # "K",
        ],
        index=pd.MultiIndex.from_tuples(
            [
                ("ma", "Cold"),
                # ("mb", "Cold"),
            ],
            names=["model", "variable"],
        ),
    )

    with pytest.raises(MissingDesiredUnitError):
        convert_unit_from_target_series(start, desired_unit)


@pobj_type
def test_convert_unit_from_target_series_no_pint_error(pobj_type):
    start = pd.DataFrame(
        np.arange(2 * 3).reshape((2, 3)),
        columns=np.array([1.0, 10.0, 100.0]),
        index=pd.MultiIndex.from_tuples(
            [
                ("ma", "Cold", "microK"),
                ("mb", "Cold", "K"),
            ],
            names=["model", "variable", "unit"],
        ),
    )

    start = convert_to_desired_type(start, pobj_type)

    desired_unit = pd.Series(
        ["K", "K"],
        index=pd.MultiIndex.from_tuples(
            [
                ("ma", "Cold"),
                ("mb", "Cold"),
            ],
            names=["model", "variable"],
        ),
    )

    with patch.dict(sys.modules, {"pint": None}):
        with pytest.raises(
            MissingOptionalDependencyError,
            match=re.escape(
                "`convert_unit_from_target_series(..., ur=None, ...)` "
                "requires pint to be installed"
            ),
        ):
            convert_unit_from_target_series(start, desired_unit)


@pobj_type
def test_accessor_convert_unit(setup_pandas_accessors, pobj_type):
    # Do most complex case: supply a series with different unit level
    # and required unit registry
    openscm_units = pytest.importorskip("openscm_units")

    start = create_test_df(
        variables=[
            ("temperature", "K"),
            ("erf", "W / m^2"),
            ("ohc", "ZJ"),
            ("emissions", "GtC"),
        ],
        n_scenarios=2,
        n_runs=2,
        timepoints=np.array([1850.0, 2000.0, 2050.0, 2100.0]),
    ).rename_axis(index={"unit": "units"})

    start = convert_to_desired_type(start, pobj_type)

    desired_units = (
        start.loc[start.index.get_level_values("variable") != "temperature"]
        .reset_index("units")["units"]
        .replace({"W / m^2": "ZJ / yr / m^2", "ZJ": "PJ", "GtC": "MtCO2"})
    )

    res = start.openscm.convert_unit(
        desired_units, unit_level="units", ur=openscm_units.unit_registry
    )

    np.testing.assert_allclose(
        res.loc[res.index.get_level_values("variable") == "temperature", :].values,
        start.loc[start.index.get_level_values("variable") == "temperature", :].values,
    )

    np.testing.assert_allclose(
        res.loc[res.index.get_level_values("variable") == "erf", :].values,
        (60.0 * 60.0 * 24.0 * 365.25)
        * 1e-21
        * start.loc[start.index.get_level_values("variable") == "erf", :].values,
    )

    np.testing.assert_allclose(
        res.loc[res.index.get_level_values("variable") == "ohc", :].values,
        1e6 * start.loc[start.index.get_level_values("variable") == "ohc", :].values,
    )

    np.testing.assert_allclose(
        res.loc[res.index.get_level_values("variable") == "emissions", :].values,
        44.0
        / 12.0
        * 1000.0
        * start.loc[start.index.get_level_values("variable") == "emissions", :].values,
    )


@pobj_type
@target_type
def test_accessor_convert_unit_like(setup_pandas_accessors, pobj_type, target_type):
    # Do most complex case: supply a series with different unit level
    # and required unit registry
    openscm_units = pytest.importorskip("openscm_units")

    start = create_test_df(
        variables=[
            ("temperature", "K"),
            ("erf", "W / m^2"),
            ("ohc", "ZJ"),
            ("emissions", "GtC"),
        ],
        n_scenarios=2,
        n_runs=2,
        timepoints=np.array([1850.0, 2000.0, 2050.0, 2100.0]),
    ).rename_axis(index={"unit": "units"})

    start = convert_to_desired_type(start, pobj_type)

    target = create_test_df(
        variables=[
            ("temperature", "mK"),
            ("erf", "W / m^2"),
            ("ohc", "PJ"),
            ("emissions", "MtC"),
        ],
        n_scenarios=2,
        n_runs=2,
        timepoints=np.array([1850.0, 2000.0, 2050.0, 2100.0]),
    ).rename_axis(index={"unit": "unit_level"})

    res = start.openscm.convert_unit_like(
        target,
        unit_level="units",
        target_unit_level="unit_level",
        ur=openscm_units.unit_registry,
    )

    target = convert_to_desired_type(target, target_type)

    exp = start.openscm.convert_unit(
        {"K": "mK", "ZJ": "PJ", "GtC": "MtC"},
        unit_level="units",
        ur=openscm_units.unit_registry,
    )

    check_result(res, exp)
